# Copyright 2013 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from tests.unit import base
from chirp import chirp_common
from chirp.drivers import icf


class TestBaseMapping(base.BaseTest):
    CLS = chirp_common.MemoryMapping

    def test_mapping(self):
        model = chirp_common.MappingModel(None, 'Foo')
        mapping = self.CLS(model, 1, 'Foo')
        self.assertEqual(str(mapping), 'Foo')
        self.assertEqual(mapping.get_name(), 'Foo')
        self.assertEqual(mapping.get_index(), 1)
        self.assertEqual(repr(mapping), '%s-1' % self.CLS.__name__)
        self.assertEqual(mapping._model, model)

    def test_mapping_eq(self):
        mapping1 = self.CLS(None, 1, 'Foo')
        mapping2 = self.CLS(None, 1, 'Bar')
        mapping3 = self.CLS(None, 2, 'Foo')

        self.assertEqual(mapping1, mapping2)
        self.assertNotEqual(mapping1, mapping3)


class TestBaseBank(TestBaseMapping):
    CLS = chirp_common.Bank


class _TestBaseClass(base.BaseTest):
    ARGS = tuple()

    def setUp(self):
        super(_TestBaseClass, self).setUp()
        self.model = self.CLS(*self.ARGS)

    def _test_base(self, method, *args):
        self.assertRaises(NotImplementedError,
                          getattr(self.model, method), *args)


class TestBaseMappingModel(_TestBaseClass):
    CLS = chirp_common.MappingModel
    ARGS = tuple([None, 'Foo'])

    def test_base_class(self):
        methods = [('get_num_mappings', ()),
                   ('get_mappings', ()),
                   ('add_memory_to_mapping', (None, None)),
                   ('remove_memory_from_mapping', (None, None)),
                   ('get_mapping_memories', (None,)),
                   ('get_memory_mappings', (None,)),
                   ]
        for method, args in methods:
            self._test_base(method, *args)

    def test_get_name(self):
        self.assertEqual(self.model.get_name(), 'Foo')


class TestBaseBankModel(TestBaseMappingModel):
    ARGS = tuple([None])
    CLS = chirp_common.BankModel

    def test_get_name(self):
        self.assertEqual(self.model.get_name(), 'Banks')


class TestBaseMappingModelIndexInterface(_TestBaseClass):
    CLS = chirp_common.MappingModelIndexInterface

    def test_base_class(self):
        methods = [('get_index_bounds', ()),
                   ('get_memory_index', (None, None)),
                   ('set_memory_index', (None, None, None)),
                   ('get_next_mapping_index', (None,)),
                   ]
        for method, args in methods:
            self._test_base(method, *args)


class TestIcomBanks(TestBaseMapping):
    def test_icom_bank(self):
        bank = icf.IcomBank(None, 1, 'Foo')
        # IcomBank has an index attribute used by IcomBankModel
        self.assertTrue(hasattr(bank, 'index'))


class TestIcomBankModel(base.BaseTest):
    CLS = icf.IcomBankModel

    def _get_rf(self):
        rf = chirp_common.RadioFeatures()
        rf.memory_bounds = (1, 10)
        return rf

    def setUp(self):
        super(TestIcomBankModel, self).setUp()

        class FakeRadio(icf.IcomCloneModeRadio):
            _num_banks = 10
            _bank_index_bounds = (0, 10)

            def get_features(the_radio):
                return self._get_rf()

            def _set_bank(self, number, index):
                pass

            def _get_bank(self, number):
                pass

            def _get_bank_index(self, number):
                pass

            def _set_bank_index(self, number, index):
                pass

            def get_memory(self, number):
                pass

        self._radio = FakeRadio(None)
        self._model = self.CLS(self._radio)
        for a in ('_set_bank', '_get_bank', '_set_bank_index',
                  '_get_bank_index', 'get_memory'):
            m = mock.patch.object(self._radio, a)
            m.start()
            self.mocks.append(m)

    def test_get_num_mappings(self):
        self.assertEqual(self._model.get_num_mappings(), 10)

    def test_get_mappings(self):
        banks = self._model.get_mappings()
        self.assertEqual(len(banks), 10)
        i = 0
        for bank in banks:
            index = chr(ord("A") + i)
            self.assertEqual(bank.get_index(), index)
            self.assertEqual(bank.get_name(), 'BANK-%s' % index)
            self.assertEqual(bank.index, i)
            i += 1

    def test_add_memory_to_mapping(self):
        mem = chirp_common.Memory()
        mem.number = 5
        banks = self._model.get_mappings()
        bank = banks[2]
        self._model.add_memory_to_mapping(mem, bank)
        self._radio._set_bank.assert_called_once_with(5, 2)

    def _setup_test_remove_memory_from_mapping(self, curbank):
        mem = chirp_common.Memory()
        mem.number = 5
        banks = self._model.get_mappings()
        bank = banks[2]
        self._radio._get_bank.return_value = curbank
        return mem, bank

    def test_remove_memory_from_mapping(self):
        mem, bank = self._setup_test_remove_memory_from_mapping(2)
        self._model.remove_memory_from_mapping(mem, bank)
        self._radio._get_bank.assert_called_once_with(5)
        self._radio._set_bank.assert_called_once_with(5, None)

    def test_remove_memory_from_mapping_wrong_bank(self):
        mem, bank = self._setup_test_remove_memory_from_mapping(3)
        self.assertRaises(Exception,
                          self._model.remove_memory_from_mapping, mem, bank)

    def test_remove_memory_from_mapping_no_bank(self):
        mem, bank = self._setup_test_remove_memory_from_mapping(None)
        self.assertRaises(Exception,
                          self._model.remove_memory_from_mapping, mem, bank)

    def test_get_mapping_memories(self):
        banks = self._model.get_mappings()
        expected = []
        _get_bank = []
        for i in range(1, 10):
            should_include = bool(i % 2)
            _get_bank.append(
                should_include and banks[1].index or None)
            if should_include:
                expected.append(i)
        self._radio._get_bank.side_effect = _get_bank
        self._radio.get_memory.side_effect = expected
        members = self._model.get_mapping_memories(banks[1])
        self.assertEqual(members, expected)
        self._radio.get_memory.assert_has_calls([
            mock.call(i) for i in range(1, 10) if i % 2])
        self._radio._get_bank.assert_has_calls([
            mock.call(i) for i in range(1, 10)])

    def test_get_memory_mappings(self):
        banks = self._model.get_mappings()
        mem1 = chirp_common.Memory()
        mem1.number = 5
        mem2 = chirp_common.Memory()
        mem2.number = 6
        self._radio._get_bank.side_effect = [2, None]
        self.assertEqual(self._model.get_memory_mappings(mem1)[0], banks[2])
        self.assertEqual(self._model.get_memory_mappings(mem2), [])
        self._radio._get_bank.assert_has_calls([
            mock.call(mem1.number), mock.call(mem2.number)])


class TestIcomIndexedBankModel(TestIcomBankModel):
    CLS = icf.IcomIndexedBankModel

    def _get_rf(self):
        rf = super(TestIcomIndexedBankModel, self)._get_rf()
        rf.has_bank_index = True
        return rf

    def test_get_index_bounds(self):
        self.assertEqual(self._model.get_index_bounds(), (0, 10))

    def test_get_memory_index(self):
        mem = chirp_common.Memory()
        mem.number = 5
        self._radio._get_bank_index.return_value = 1
        self.assertEqual(self._model.get_memory_index(mem, None), 1)
        self._radio._get_bank_index.assert_called_once_with(mem.number)

    def test_set_memory_index(self):
        mem = chirp_common.Memory()
        mem.number = 5
        banks = self._model.get_mappings()
        with mock.patch.object(self._model, 'get_memory_mappings') as m:
            m.return_value = [banks[3]]
            self._model.set_memory_index(mem, banks[3], 1)
            m.assert_called_once_with(mem)
        self._radio._set_bank_index.assert_called_once_with(mem.number, 1)

    def test_set_memory_index_bad_bank(self):
        mem = chirp_common.Memory()
        mem.number = 5
        banks = self._model.get_mappings()
        with mock.patch.object(self._model, 'get_memory_mappings') as m:
            m.return_value = [banks[4]]
            self.assertRaises(Exception,
                              self._model.set_memory_index, mem, banks[3], 1)
            m.assert_called_once_with(mem)

    def test_set_memory_index_bad_index(self):
        mem = chirp_common.Memory()
        mem.number = 5
        banks = self._model.get_mappings()
        with mock.patch.object(self._model, 'get_memory_mappings') as m:
            m.return_value = [banks[3]]
            self.assertRaises(Exception,
                              self._model.set_memory_index, mem, banks[3], 99)
            m.assert_called_once_with(mem)

    def test_get_next_mapping_index(self):
        banks = self._model.get_mappings()
        bank_exp_calls = []
        index_exp_calls = []
        bank_returns = []
        index_returns = []
        for i in range(*self._radio.get_features().memory_bounds):
            bank_returns.append((i % 2) and banks[1].index)
            bank_exp_calls.append(mock.call(i))
            if bool(i % 2):
                index_returns.append(i)
                index_exp_calls.append(mock.call(i))
        idx = 0
        for i in range(*self._radio.get_features().memory_bounds):
            bank_exp_calls.append(mock.call(i))
            bank_returns.append((i % 2) and banks[2].index)
            if i % 2:
                index_exp_calls.append(mock.call(i))
                index_returns.append(idx)
                idx += 1

        self._radio._get_bank.side_effect = bank_returns
        self._radio._get_bank_index.side_effect = index_returns
        self.assertEqual(self._model.get_next_mapping_index(banks[1]), 0)
        self.assertEqual(self._model.get_next_mapping_index(banks[2]), 5)
        self._radio._get_bank.assert_has_calls(bank_exp_calls)
        self._radio._get_bank_index.assert_has_calls(index_exp_calls)



import os

import pytest


def pytest_collection_modifyitems(config, items):
    xfails = os.path.join(os.path.dirname(__file__), 'xfails.txt')
    with open(xfails) as f:
        lines = [x.strip() for x in f.readlines() if not x.startswith('#')]
    msg = ('Test marked for XFAIL in tests/xfails.txt, but did not fail. '
           'If this test is now fixed, remove it from the file.')
    for item in items:
        if item.nodeid in lines:
            mark = pytest.mark.xfail(reason=msg)
            item.add_marker(mark)



# Copyright 2019 Dan Smith <dsmith@danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
from builtins import bytes

import struct

from chirp.drivers import icf

import logging

LOG = logging.getLogger(__name__)


log = logging.getLogger()
log.level = logging.ERROR
log.addHandler(logging.StreamHandler(sys.stdout))


class FakeIcomRadio(object):
    def __init__(self, radio, mapfile=None):
        self._buffer = bytes(b'')
        self._radio = radio
        if not mapfile:
            self._memory = bytes(b'\x00') * radio.get_memsize()
        else:
            self.load_from_file(mapfile)

    def load_from_file(self, filename):
        with open(filename, 'rb') as f:
            self._memory = bytes(f.read())
        LOG.debug('Initialized %i bytes from %s' % (len(self._memory),
                                                    filename))

    def read(self, count):
        """read() from radio, so here we synthesize responses"""
        chunk = self._buffer[:count]
        self._buffer = self._buffer[count:]
        return chunk

    def queue(self, data):
        # LOG.debug('Queuing: %r' % data)
        self._buffer += data

    def make_response(self, cmd, payload):
        return bytes([
            0xFE, 0xFE,
            0xEF,  # Radio
            0xEE,  # PC
            cmd,
        ]) + payload + bytes([0xFD])

    @property
    def address_fmt(self):
        if self._radio.get_memsize() > 0x10000:
            return 'I'
        else:
            return 'H'

    def do_clone_out(self):
        LOG.debug('Clone from radio started')
        size = 16
        for addr in range(0, self._radio.get_memsize(), size):
            if len(self._memory[addr:]) < 4:
                # IC-W32E has an off-by-one hack for detection,
                # which will cause us to send a short one-byte
                # block of garbage, unlike the real radio. So,
                # if we get to the end and have a few bytes
                # left, don't be stupid.
                break
            header = bytes(struct.pack('>%sB' % self.address_fmt,
                                       addr, size))
            chunk = []
            cs = 0
            for byte in header:
                chunk.extend(x for x in bytes(b'%02X' % byte))
                cs += byte
            for byte in self._memory[addr:addr + size]:
                chunk.extend(x for x in bytes(b'%02X' % byte))
                cs += byte

            vx = ((cs ^ 0xFFFF) + 1) & 0xFF
            chunk.extend(x for x in bytes(b'%02X' % vx))
            self.queue(self.make_response(icf.CMD_CLONE_DAT, bytes(chunk)))
        self.queue(self.make_response(icf.CMD_CLONE_END, bytes([])))

    def do_clone_in(self):
        LOG.debug('Clone to radio started')
        self._memory = bytes(b'')

    def do_clone_data(self, payload_hex):
        if self.address_fmt == 'I':
            header_len = 5
        else:
            header_len = 3

        def hex_to_byte(hexchars):
            return int('%s%s' % (chr(hexchars[0]), chr(hexchars[1])), 16)

        payload_bytes = bytes([hex_to_byte(payload_hex[i:i+2])
                               for i in range(0, len(payload_hex), 2)])

        addr, size = struct.unpack(
            '>%sB' % self.address_fmt, payload_bytes[:header_len])
        data = payload_bytes[header_len:-1]
        # csum = payload_bytes[-1]

        if len(data) != size:
            LOG.debug('Invalid frame size: expected %i, but got %i' % (
                size, len(data)))

        expected_addr = len(self._memory)
        if addr < expected_addr:
            LOG.debug('Frame goes back to %04x from %04x' % (addr,
                                                             expected_addr))
        if len(self._memory) != addr:
            LOG.debug('Filling gap between %04x and %04x' % (expected_addr,
                                                             addr))
            self._memory += (bytes(b'\x00') * (addr - expected_addr))

        # FIXME: Check checksum

        self._memory += data

    def do_clone_end(self):
        self.queue(self.make_response(icf.CMD_CLONE_OK, bytes([0])))

    def write(self, data):
        """write() to radio, so here we process requests"""

        assert isinstance(
            data, bytes), 'Bytes required, %s received' % data.__class__

        while data.startswith(b'\xfe\xfe\xfe'):
            data = data[1:]

        # src = data[2]
        # dst = data[3]
        cmd = data[4]
        payload = data[5:-1]
        # end = data[-1]

        LOG.debug('Received command: %x' % cmd)
        LOG.debug('  Full frame: %r' % data)

        model = self._radio.get_model() + bytes(b'\x00' * 20)

        if cmd == 0xE0:  # Ident
            # FIXME
            self.queue(self.make_response(0x01,  # Model
                                          model))
        elif cmd == icf.CMD_CLONE_OUT:
            self.do_clone_out()
        elif cmd == icf.CMD_CLONE_IN:
            self.do_clone_in()
        elif cmd == icf.CMD_CLONE_DAT:
            self.do_clone_data(payload)
        elif cmd == icf.CMD_CLONE_END:
            self.do_clone_end()
        elif cmd == icf.CMD_CLONE_HISPEED:
            LOG.info('Got hispeed kicker')
        else:
            LOG.debug('Unknown command %i' % cmd)
            self.queue(self.make_response(0x00, bytes([0x01])))

        return len(data)

    def flush(self):
        return



import base64
import copy
import json
import os
import pickle
import tempfile

from unittest import mock

from tests.unit import base
from chirp import CHIRP_VERSION
from chirp import chirp_common
from chirp import directory
from chirp import errors
from chirp import settings


class TestUtilityFunctions(base.BaseTest):
    def test_parse_freq_whole(self):
        self.assertEqual(chirp_common.parse_freq("146.520000"), 146520000)
        self.assertEqual(chirp_common.parse_freq("146.5200"), 146520000)
        self.assertEqual(chirp_common.parse_freq("146.52"), 146520000)
        self.assertEqual(chirp_common.parse_freq("146"), 146000000)
        self.assertEqual(chirp_common.parse_freq("1250"), 1250000000)
        self.assertEqual(chirp_common.parse_freq("123456789"),
                         123456789000000)

    def test_parse_freq_decimal(self):
        self.assertEqual(chirp_common.parse_freq("1.0"), 1000000)
        self.assertEqual(chirp_common.parse_freq("1.000000"), 1000000)
        self.assertEqual(chirp_common.parse_freq("1.1"), 1100000)
        self.assertEqual(chirp_common.parse_freq("1.100"), 1100000)
        self.assertEqual(chirp_common.parse_freq("0.6"), 600000)
        self.assertEqual(chirp_common.parse_freq("0.600"), 600000)
        self.assertEqual(chirp_common.parse_freq("0.060"), 60000)
        self.assertEqual(chirp_common.parse_freq(".6"), 600000)

    def test_parse_freq_whitespace(self):
        self.assertEqual(chirp_common.parse_freq("1  "), 1000000)
        self.assertEqual(chirp_common.parse_freq("   1"), 1000000)
        self.assertEqual(chirp_common.parse_freq("   1  "), 1000000)

        self.assertEqual(chirp_common.parse_freq("1.0  "), 1000000)
        self.assertEqual(chirp_common.parse_freq("   1.0"), 1000000)
        self.assertEqual(chirp_common.parse_freq("   1.0  "), 1000000)
        self.assertEqual(chirp_common.parse_freq(""), 0)
        self.assertEqual(chirp_common.parse_freq(" "), 0)

    def test_parse_freq_bad(self):
        self.assertRaises(ValueError, chirp_common.parse_freq, "a")
        self.assertRaises(ValueError, chirp_common.parse_freq, "1.a")
        self.assertRaises(ValueError, chirp_common.parse_freq, "a.b")
        self.assertRaises(ValueError, chirp_common.parse_freq,
                          "1.0000001")

    def test_format_freq(self):
        self.assertEqual(chirp_common.format_freq(146520000), "146.520000")
        self.assertEqual(chirp_common.format_freq(54000000), "54.000000")
        self.assertEqual(chirp_common.format_freq(1800000), "1.800000")
        self.assertEqual(chirp_common.format_freq(1), "0.000001")
        self.assertEqual(chirp_common.format_freq(1250000000), "1250.000000")

    @mock.patch('chirp.CHIRP_VERSION', new='daily-20151021')
    def test_compare_version_to_current(self):
        self.assertTrue(chirp_common.is_version_newer('daily-20180101'))
        self.assertFalse(chirp_common.is_version_newer('daily-20140101'))
        self.assertFalse(chirp_common.is_version_newer('0.3.0'))
        self.assertFalse(chirp_common.is_version_newer('0.3.0dev'))

    @mock.patch('chirp.CHIRP_VERSION', new='0.3.0dev')
    def test_compare_version_to_current_dev(self):
        self.assertTrue(chirp_common.is_version_newer('daily-20180101'))

    def test_from_Hz(self):
        # FIXME: These are wrong! Adding them here purely to test the
        # python3 conversion, but they should be fixed.
        self.assertEqual(140, chirp_common.from_GHz(14000000001))
        self.assertEqual(140, chirp_common.from_MHz(14000001))
        self.assertEqual(140, chirp_common.from_kHz(14001))

    def test_mem_to_from_csv(self):
        mem1 = chirp_common.Memory()
        mem2 = chirp_common.Memory._from_csv(','.join(mem1.to_csv()))
        self.assertEqual(str(mem1), str(mem2))

    def test_mem_from_text_rb1(self):
        text = '145.2500 	-0.6 MHz 	97.4 	OPEN'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(145250000, mem.freq)
        self.assertEqual(600000, mem.offset)
        self.assertEqual('-', mem.duplex)
        self.assertEqual('Tone', mem.tmode)
        self.assertEqual(97.4, mem.rtone)

    def test_mem_from_text_rb2(self):
        text = '147.3000 	+0.6 MHz 	156.7 / 156.7 	OPEN'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(147300000, mem.freq)
        self.assertEqual(600000, mem.offset)
        self.assertEqual('+', mem.duplex)
        self.assertEqual('TSQL', mem.tmode)
        self.assertEqual(156.7, mem.ctone)

    def test_mem_from_text_rb3(self):
        text = '441.1000 	+5 MHz 	D125 / D125 	OPEN'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(441100000, mem.freq)
        self.assertEqual(5000000, mem.offset)
        self.assertEqual('+', mem.duplex)
        self.assertEqual('DTCS', mem.tmode)
        self.assertEqual(125, mem.dtcs)

    def test_mem_from_text_rb4(self):
        text = '441.1000 	+5 MHz 	88.5 / D125 	OPEN'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(441100000, mem.freq)
        self.assertEqual(5000000, mem.offset)
        self.assertEqual('+', mem.duplex)
        self.assertEqual('Cross', mem.tmode)
        self.assertEqual(88.5, mem.rtone)
        self.assertEqual(125, mem.rx_dtcs)
        self.assertEqual('Tone->DTCS', mem.cross_mode)

    def test_mem_from_text_random1(self):
        text = 'Glass Butte 147.200 + 162.2'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(147200000, mem.freq)
        # This is a default
        self.assertEqual(600000, mem.offset)
        # Just a random + or - on the line isn't enough to trigger offset
        # pattern
        self.assertEqual('+', mem.duplex)
        self.assertEqual('Tone', mem.tmode)
        self.assertEqual(162.2, mem.rtone)

    def test_mem_from_text_random2(self):
        text = 'Glass - Butte 147.200 + 162.2'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(147200000, mem.freq)
        # This is a default
        self.assertEqual(600000, mem.offset)
        # Just a random + or - on the line isn't enough to trigger offset
        # pattern
        self.assertEqual('+', mem.duplex)
        self.assertEqual('Tone', mem.tmode)
        self.assertEqual(162.2, mem.rtone)

    def test_mem_from_text_random3(self):
        text = 'Glass - Butte 147.200 + 88.5'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(147200000, mem.freq)
        # This is a default
        self.assertEqual(600000, mem.offset)
        # Just a random + or - on the line isn't enough to trigger offset
        # pattern
        self.assertEqual('+', mem.duplex)
        self.assertEqual('Tone', mem.tmode)
        self.assertEqual(88.5, mem.rtone)

    def test_mem_from_text_random4(self):
        text = '146.640 | 146.040 | 136.5'
        mem = chirp_common.mem_from_text(text)
        self.assertIsNotNone(mem)
        self.assertEqual(146640000, mem.freq)
        self.assertEqual(600000, mem.offset)
        self.assertEqual('-', mem.duplex)
        self.assertEqual('Tone', mem.tmode)
        self.assertEqual(136.5, mem.rtone)

    def test_mem_from_text_chirp1(self):
        text = '[400.625000/136.000000]'
        chirp_common.mem_from_text(text)

    def test_mem_from_text_chirp2(self):
        text = '[462.675000/+5.000/136.5/136.5]'
        mem = chirp_common.mem_from_text(text)
        self.assertEqual(462675000, mem.freq)
        self.assertEqual('+', mem.duplex)
        self.assertEqual(5000000, mem.offset)

    def test_mem_from_text_chirp3(self):
        text = '[1.675000/1.685/136.5/136.5]'
        mem = chirp_common.mem_from_text(text)
        self.assertEqual(1675000, mem.freq)
        self.assertEqual('+', mem.duplex)
        self.assertEqual(10000, mem.offset)

    def test_mem_from_text_chirp4(self):
        text = '[450.000000/150.000000]'
        mem = chirp_common.mem_from_text(text)
        self.assertEqual(450000000, mem.freq)
        self.assertEqual(150000000, mem.offset)
        self.assertEqual('split', mem.duplex)

    def test_mem_from_text_chirp5(self):
        text = '[500.000000/-9.900]'
        mem = chirp_common.mem_from_text(text)
        self.assertEqual(500000000, mem.freq)
        self.assertEqual(9900000, mem.offset)
        self.assertEqual('-', mem.duplex)
        self.assertEqual(text, chirp_common.mem_to_text(mem))

    def test_mem_from_text_chirp6(self):
        text = '[450.000000/150.000]'
        mem = chirp_common.mem_from_text(text)
        self.assertEqual(450000000, mem.freq)
        self.assertEqual(150000000, mem.offset)
        self.assertEqual('split', mem.duplex)

    def test_mem_from_text_chirp7(self):
        # Offsets >= 10MHz are not allowed, so this will get
        # parsed as a tx frequency of 15MHz
        text = '[450.000000/+15.000]'
        mem = chirp_common.mem_from_text(text)
        self.assertEqual(450000000, mem.freq)
        self.assertEqual(15000000, mem.offset)
        self.assertEqual('split', mem.duplex)

    def test_mem_from_text_chirp8(self):
        # Offsets >= 10MHz are not allowed, so this will get
        # parsed as a tx frequency of 15MHz
        text = '[450.000000/+150.000]'
        mem = chirp_common.mem_from_text(text)
        self.assertEqual(450000000, mem.freq)
        self.assertEqual(150000000, mem.offset)
        self.assertEqual('split', mem.duplex)

    def test_mem_from_text_tsv(self):
        from_lo = """8	Mt Scott	147.28	+	0.6	TSQL	88.5	167.9	23	NN	23	Tone->Tone	FM	5		50W	TayDan Emergency Step #2
9	Mt Tabor	145.39	-	0.6	TSQL	88.5	100	23	NN	23	Tone->Tone	FM	5		50W	TayDan Emergency Step #1
10	Timberline V	147.12	+	0.6	TSQL	103.5	100	23	NN	23	Tone->Tone	FM	5		50W	TayDan Emergency Step #4
11	Timberline U	444.225	+	5	TSQL	88.5	100	23	NN	23	Tone->Tone	FM	5		50W	K7RPT Timberline, Linked to VHF side, AC7QE-R echolink
12	GovtCamp	443.875	+	5	TSQL	88.5	103.5	23	NN	23	Tone->Tone	FM	5		50W"""  # noqa

        lines = from_lo.split('\n')
        mem = chirp_common.mem_from_text(lines[0])
        self.assertEqual('Mt Scott', mem.name)
        self.assertEqual('TSQL', mem.tmode)
        self.assertEqual('+', mem.duplex)
        self.assertEqual(600000, mem.offset)
        self.assertEqual(147280000, mem.freq)
        self.assertEqual(8, mem.number)
        for line in from_lo.split('\n'):
            chirp_common.mem_from_text(line)

    def test_mem_to_text1(self):
        mem = chirp_common.Memory()
        mem.freq = 146900000
        mem.duplex = '-'
        mem.offset = 600000
        mem.tmode = 'TSQL'
        mem.ctone = 100.0
        txt = chirp_common.mem_to_text(mem)
        self.assertEqual('[146.900000/-0.600/100.0/100.0]', txt)
        chirp_common.mem_from_text(txt)
        self.assertEqual(600000, mem.offset)
        self.assertEqual('-', mem.duplex)

    def test_mem_to_text2(self):
        mem = chirp_common.Memory()
        mem.freq = 146900000
        mem.duplex = 'split'
        mem.offset = 446000000
        mem.tmode = 'Cross'
        mem.cross_mode = 'Tone->DTCS'
        mem.rtone = 100.0
        mem.rx_dtcs = 25
        txt = chirp_common.mem_to_text(mem)
        self.assertEqual('[146.900000/446.000000/100.0/D025]', txt)
        chirp_common.mem_from_text(txt)

    def test_mem_to_text3(self):
        mem = chirp_common.Memory()
        mem.freq = 146520000
        mem.duplex = ''
        mem.offset = 600000
        mem.tmode = 'DTCS'
        mem.dtcs = 25
        txt = chirp_common.mem_to_text(mem)
        self.assertEqual('[146.520000/D025/D025]', txt)
        chirp_common.mem_from_text(txt)

    def test_parse_power(self):
        valid = [
            ('0.1', 0.1),
            ('0.1W', 0.1),
            ('0.25', 0.2),
            ('0.25W', 0.2),
            ('11.0', 11),
            ('11', 11),
            ('11.0W', 11),
            ('11w', 11),
            ('1500', 1500),
            ('2500', 2500),
            ('2500W', 2500),
            ('2500.0W', 2500)]
        for s, v in valid:
            power = chirp_common.parse_power(s)
            self.assertEqual(v, chirp_common.dBm_to_watts(float(power)))

    def test_parse_power_invalid(self):
        invalid = ['2500d', '2d1', 'aaa', 'a', '']
        for s in invalid:
            self.assertRaises(ValueError, chirp_common.parse_power, s)

    def test_airband(self):
        self.assertTrue(chirp_common.is_airband(108000000))
        self.assertTrue(chirp_common.is_airband(108000000 + 25000))
        self.assertTrue(chirp_common.is_airband(136000000 - 25000))
        self.assertTrue(chirp_common.is_airband(136000000))

        self.assertFalse(chirp_common.is_airband(108000000 - 1))
        self.assertFalse(chirp_common.is_airband(137000000 + 1))


class TestSplitTone(base.BaseTest):
    def _test_split_tone_decode(self, tx, rx, **vals):
        mem = chirp_common.Memory()
        chirp_common.split_tone_decode(mem, tx, rx)
        for key, value in list(vals.items()):
            self.assertEqual(getattr(mem, key), value)

    def test_split_tone_decode_none(self):
        self._test_split_tone_decode((None, None, None),
                                     (None, None, None),
                                     tmode='')

    def test_split_tone_decode_tone(self):
        self._test_split_tone_decode(('Tone', 100.0, None),
                                     ('', 0, None),
                                     tmode='Tone',
                                     rtone=100.0)

    def test_split_tone_decode_tsql(self):
        self._test_split_tone_decode(('Tone', 100.0, None),
                                     ('Tone', 100.0, None),
                                     tmode='TSQL',
                                     ctone=100.0)

    def test_split_tone_decode_dtcs(self):
        self._test_split_tone_decode(('DTCS', 23, None),
                                     ('DTCS', 23, None),
                                     tmode='DTCS',
                                     dtcs=23)

    def test_split_tone_decode_cross_tone_tone(self):
        self._test_split_tone_decode(('Tone', 100.0, None),
                                     ('Tone', 123.0, None),
                                     tmode='Cross',
                                     cross_mode='Tone->Tone',
                                     rtone=100.0,
                                     ctone=123.0)

    def test_split_tone_decode_cross_tone_dtcs(self):
        self._test_split_tone_decode(('Tone', 100.0, None),
                                     ('DTCS', 32, 'R'),
                                     tmode='Cross',
                                     cross_mode='Tone->DTCS',
                                     rtone=100.0,
                                     rx_dtcs=32,
                                     dtcs_polarity='NR')

    def test_split_tone_decode_cross_dtcs_tone(self):
        self._test_split_tone_decode(('DTCS', 32, 'R'),
                                     ('Tone', 100.0, None),
                                     tmode='Cross',
                                     cross_mode='DTCS->Tone',
                                     ctone=100.0,
                                     dtcs=32,
                                     dtcs_polarity='RN')

    def test_split_tone_decode_cross_dtcs_dtcs(self):
        self._test_split_tone_decode(('DTCS', 32, 'R'),
                                     ('DTCS', 25, 'R'),
                                     tmode='Cross',
                                     cross_mode='DTCS->DTCS',
                                     dtcs=32,
                                     rx_dtcs=25,
                                     dtcs_polarity='RR')

    def test_split_tone_decode_cross_none_dtcs(self):
        self._test_split_tone_decode((None, None, None),
                                     ('DTCS', 25, 'R'),
                                     tmode='Cross',
                                     cross_mode='->DTCS',
                                     rx_dtcs=25,
                                     dtcs_polarity='NR')

    def test_split_tone_decode_cross_none_tone(self):
        self._test_split_tone_decode((None, None, None),
                                     ('Tone', 100.0, None),
                                     tmode='Cross',
                                     cross_mode='->Tone',
                                     ctone=100.0)

    def _set_mem(self, **vals):
        mem = chirp_common.Memory()
        for key, value in list(vals.items()):
            setattr(mem, key, value)
        return chirp_common.split_tone_encode(mem)

    def split_tone_encode_test_none(self):
        self.assertEqual(self._set_mem(tmode=''),
                         (('', None, None),
                          ('', None, None)))

    def split_tone_encode_test_tone(self):
        self.assertEqual(self._set_mem(tmode='Tone', rtone=100.0),
                         (('Tone', 100.0, None),
                          ('', None, None)))

    def split_tone_encode_test_tsql(self):
        self.assertEqual(self._set_mem(tmode='TSQL', ctone=100.0),
                         (('Tone', 100.0, None),
                          ('Tone', 100.0, None)))

    def split_tone_encode_test_dtcs(self):
        self.assertEqual(self._set_mem(tmode='DTCS', dtcs=23,
                                       dtcs_polarity='RN'),
                         (('DTCS', 23, 'R'),
                          ('DTCS', 23, 'N')))

    def split_tone_encode_test_cross_tone_tone(self):
        self.assertEqual(self._set_mem(tmode='Cross', cross_mode='Tone->Tone',
                                       rtone=100.0, ctone=123.0),
                         (('Tone', 100.0, None),
                          ('Tone', 123.0, None)))

    def split_tone_encode_test_cross_tone_dtcs(self):
        self.assertEqual(self._set_mem(tmode='Cross', cross_mode='Tone->DTCS',
                                       rtone=100.0, rx_dtcs=25),
                         (('Tone', 100.0, None),
                          ('DTCS', 25, 'N')))

    def split_tone_encode_test_cross_dtcs_tone(self):
        self.assertEqual(self._set_mem(tmode='Cross', cross_mode='DTCS->Tone',
                                       ctone=100.0, dtcs=25),
                         (('DTCS', 25, 'N'),
                          ('Tone', 100.0, None)))

    def split_tone_encode_test_cross_none_dtcs(self):
        self.assertEqual(self._set_mem(tmode='Cross', cross_mode='->DTCS',
                                       rx_dtcs=25),
                         (('', None, None),
                          ('DTCS', 25, 'N')))

    def split_tone_encode_test_cross_none_tone(self):
        self.assertEqual(self._set_mem(tmode='Cross', cross_mode='->Tone',
                                       ctone=100.0),
                         (('', None, None),
                          ('Tone', 100.0, None)))


class TestStepFunctions(base.BaseTest):
    _625 = [145856250,
            445856250,
            862731250,
            146118750,
            ]
    _125 = [145862500,
            445862500,
            862737500,
            ]
    _005 = [145005000,
            445005000,
            850005000,
            ]
    _025 = [145002500,
            445002500,
            850002500,
            ]

    def test_is_fractional_step(self):
        for freq in self._125 + self._625:
            print(freq)
            self.assertTrue(chirp_common.is_fractional_step(freq))

    def test_is_6_25(self):
        for freq in self._625:
            self.assertTrue(chirp_common.is_6_25(freq))

    def test_is_12_5(self):
        for freq in self._125:
            self.assertTrue(chirp_common.is_12_5(freq))

    def test_is_5_0(self):
        for freq in self._005:
            self.assertTrue(chirp_common.is_5_0(freq))

    def test_is_2_5(self):
        for freq in self._025:
            self.assertTrue(chirp_common.is_2_5(freq))

    def test_required_step(self):
        steps = {2.5: self._025,
                 5.0: self._005,
                 6.25: self._625,
                 12.5: self._125,
                 }
        for step, freqs in list(steps.items()):
            for freq in freqs:
                self.assertEqual(step, chirp_common.required_step(freq))

    def test_required_step_with_list(self):
        steps = {5.0: self._005,
                 6.25: self._625,
                 12.5: self._125,
                 }
        allowed = [12.5, 6.25]
        for step, freqs in list(steps.items()):
            for freq in freqs:
                # We don't support 5.0, so any of the frequencies in that
                # list should raise an error
                if step == 5.0:
                    self.assertRaises(errors.InvalidDataError,
                                      chirp_common.required_step,
                                      freq, allowed)
                else:
                    self.assertEqual(step, chirp_common.required_step(
                        freq, allowed))

    def test_required_step_finds_suitable(self):
        # If we support 5.0, we should get it as the step for those
        self.assertEqual(5.0, chirp_common.required_step(self._005[0],
                                                         allowed=[5.0, 2.5]))
        # If we support 2.5 and not 5.0, then we should find 2.5 as a suitable
        # alternative
        self.assertEqual(2.5, chirp_common.required_step(self._005[0],
                                                         allowed=[2.5]))

    def test_required_step_finds_radio_specific(self):
        # Make sure we find a radio-specific step, 10Hz in this case
        self.assertEqual(0.01, chirp_common.required_step(
            146000010, allowed=[5.0, 10.0, 0.01, 20.0]))

    def test_required_step_fail(self):
        self.assertRaises(errors.InvalidDataError,
                          chirp_common.required_step,
                          146520500, allowed=[5.0, 10.0, 12.5])

    def test_fix_rounded_step_250(self):
        self.assertEqual(146106250,
                         chirp_common.fix_rounded_step(146106000))

    def test_fix_rounded_step_500(self):
        self.assertEqual(146112500,
                         chirp_common.fix_rounded_step(146112000))

    def test_fix_rounded_step_750(self):
        self.assertEqual(146118750,
                         chirp_common.fix_rounded_step(146118000))

    def test_fix_rounded_step_no_change(self):
        self.assertEqual(146520000,
                         chirp_common.fix_rounded_step(146520000))

    def test_fix_rounded_step_833(self):
        # 108.008 should be aligned to 8.33kHz step
        self.assertEqual(
            int(108000000 + (25000 / 3)),
            chirp_common.fix_rounded_step(chirp_common.to_MHz(108.008)))

        # 108.0 MHz should be left unchanged as it is 25kHz-aligned
        self.assertFalse(chirp_common.is_8_33(
            chirp_common.to_MHz(108)))
        self.assertEqual(chirp_common.to_MHz(108),
                         chirp_common.fix_rounded_step(
                             chirp_common.to_MHz(108)))

        # 108.025 MHz should be left unchanged
        self.assertFalse(chirp_common.is_8_33(
            chirp_common.to_MHz(108.025)))
        self.assertEqual(chirp_common.to_MHz(108.025),
                         chirp_common.fix_rounded_step(
                             chirp_common.to_MHz(108.025)))

        # Even though 146.008333 is a valid 8.33kHz step, we should
        # not align it as such since it is not in airband
        self.assertTrue(chirp_common.is_8_33(
            chirp_common.to_MHz(146.008333)))
        self.assertRaisesRegex(errors.InvalidDataError,
                               'Unable to correct.*',
                               chirp_common.fix_rounded_step,
                               chirp_common.to_MHz(146.008333))


class TestImageMetadata(base.BaseTest):
    def test_make_metadata(self):
        class TestRadio(chirp_common.CloneModeRadio):
            VENDOR = 'Dan'
            MODEL = 'Foomaster 9000'
            VARIANT = 'R'

        r = TestRadio(None)
        r._metadata['someextra'] = 'foo'
        # We should always take the base properties from the class, so this
        # should not show up in the result.
        r._metadata['vendor'] = 'oops'
        raw_metadata = r._make_metadata()
        metadata = json.loads(base64.b64decode(raw_metadata).decode())
        expected = {
            'vendor': 'Dan',
            'model': 'Foomaster 9000',
            'variant': 'R',
            'rclass': 'TestRadio',
            'chirp_version': CHIRP_VERSION,
            'someextra': 'foo',
        }
        self.assertEqual(expected, metadata)

    def test_strip_metadata(self):
        class TestRadio(chirp_common.CloneModeRadio):
            VENDOR = 'Dan'
            MODEL = 'Foomaster 9000'
            VARIANT = 'R'

        r = TestRadio(None)
        r._metadata['someextra'] = 'foo'
        raw_metadata = r._make_metadata()
        raw_data = (b'foooooooooooooooooooooo' + TestRadio.MAGIC +
                    raw_metadata)
        data, metadata = chirp_common.CloneModeRadio._strip_metadata(raw_data)
        self.assertEqual(b'foooooooooooooooooooooo', data)
        expected = {
            'vendor': 'Dan',
            'model': 'Foomaster 9000',
            'variant': 'R',
            'rclass': 'TestRadio',
            'someextra': 'foo',
            'chirp_version': CHIRP_VERSION,
        }
        self.assertEqual(expected, metadata)

    def test_load_mmap_no_metadata(self):
        fn = os.path.join(tempfile.gettempdir(), 'testfile')
        with open(fn, 'wb') as f:
            f.write(b'thisisrawdata')
            f.flush()

        with mock.patch('chirp.memmap.MemoryMapBytes.__init__') as mock_mmap:
            mock_mmap.return_value = None
            chirp_common.CloneModeRadio(None).load_mmap(fn)
            mock_mmap.assert_called_once_with(b'thisisrawdata')
        os.remove(fn)

    def test_load_mmap_bad_metadata(self):
        fn = os.path.join(tempfile.gettempdir(), 'testfile')
        with open(fn, 'wb') as f:
            f.write(b'thisisrawdata')
            f.write(chirp_common.CloneModeRadio.MAGIC + b'bad')
            f.flush()

        with mock.patch('chirp.memmap.MemoryMapBytes.__init__') as mock_mmap:
            mock_mmap.return_value = None
            chirp_common.CloneModeRadio(None).load_mmap(fn)
            mock_mmap.assert_called_once_with(b'thisisrawdata')
        os.remove(fn)

    def test_save_mmap_includes_metadata(self):
        # Make sure that a file saved with a .img extension includes
        # the metadata blob
        class TestRadio(chirp_common.CloneModeRadio):
            VENDOR = 'Dan'
            MODEL = 'Foomaster 9000'
            VARIANT = 'R'

        fn = os.path.join(tempfile.gettempdir(), 'test.img')
        r = TestRadio(None)
        r._mmap = mock.Mock()
        r._mmap.get_byte_compatible.return_value.get_packed.return_value = (
            b'thisisrawdata')
        r.save_mmap(fn)
        with open(fn, 'rb') as f:
            filedata = f.read()
        os.remove(fn)
        data, metadata = chirp_common.CloneModeRadio._strip_metadata(filedata)
        self.assertEqual(b'thisisrawdata', data)
        expected = {
            'vendor': 'Dan',
            'model': 'Foomaster 9000',
            'variant': 'R',
            'rclass': 'TestRadio',
            'chirp_version': CHIRP_VERSION,
        }
        self.assertEqual(expected, metadata)

    def test_save_mmap_no_metadata_not_img_file(self):
        # Make sure that if we save without a .img extension we do
        # not include the metadata blob
        class TestRadio(chirp_common.CloneModeRadio):
            VENDOR = 'Dan'
            MODEL = 'Foomaster 9000'
            VARIANT = 'R'

        with tempfile.NamedTemporaryFile(suffix='.txt') as f:
            fn = f.name
        r = TestRadio(None)
        r._mmap = mock.Mock()
        r._mmap.get_byte_compatible.return_value.get_packed.return_value = (
            b'thisisrawdata')
        r.save_mmap(fn)
        with open(fn, 'rb') as f:
            filedata = f.read()
        os.remove(fn)
        data, metadata = chirp_common.CloneModeRadio._strip_metadata(filedata)
        self.assertEqual(b'thisisrawdata', data)
        self.assertEqual({}, metadata)

    def test_load_mmap_saves_metadata_on_radio(self):
        class TestRadio(chirp_common.CloneModeRadio):
            VENDOR = 'Dan'
            MODEL = 'Foomaster 9000'
            VARIANT = 'R'

        with tempfile.NamedTemporaryFile(suffix='.img') as f:
            fn = f.name
        r = TestRadio(None)
        r._mmap = mock.Mock()
        r._mmap.get_byte_compatible.return_value.get_packed.return_value = (
            b'thisisrawdata')
        r.save_mmap(fn)

        newr = TestRadio(None)
        newr.load_mmap(fn)
        expected = {
            'vendor': 'Dan',
            'model': 'Foomaster 9000',
            'variant': 'R',
            'rclass': 'TestRadio',
            'chirp_version': CHIRP_VERSION,
        }
        self.assertEqual(expected, newr.metadata)

    def test_sub_devices_linked_metadata(self):
        class FakeRadio(chirp_common.CloneModeRadio):
            def get_sub_devices(self):
                return [FakeRadioSub('a'), FakeRadioSub('b')]

        class FakeRadioSub(FakeRadio):
            def __init__(self, name):
                self.VARIANT = name

        r = FakeRadio(None)
        subs = r.get_sub_devices()
        r.link_device_metadata(subs)
        subs[1]._metadata['foo'] = 'bar'
        self.assertEqual({'a': {}, 'b': {'foo': 'bar'}}, r._metadata)


class FakeRadio(chirp_common.CloneModeRadio):
    def __init__(self, *a, **k):
        super().__init__(*a, **k)
        self._mems = {}

    def get_memory(self, number):
        return self._mems[number]

    def set_memory(self, mem):
        # Simulate not storing the comment in our memory
        mem = copy.deepcopy(mem)
        mem.comment = ''
        self._mems[mem.number] = mem


class TestCloneModeExtras(base.BaseTest):
    def test_extra_comment(self):
        r = FakeRadio(None)
        m = chirp_common.Memory()
        m.number = 0
        m.freq = 146520000
        m.comment = 'a comment'
        r.set_memory(m)
        # Make sure our fake driver didn't modify our copy
        self.assertEqual('a comment', m.comment)

        m = r.get_memory(0)
        self.assertEqual(146520000, m.freq)
        # Before we call extra, we have no comment
        self.assertEqual('', m.comment)
        m = r.get_memory_extra(m)
        # We haven't called set_extra, so still nothing
        self.assertEqual('', m.comment)

        # Do a normal set, set_extra
        m.comment = 'a comment'
        r.set_memory(m)
        r.set_memory_extra(m)

        # Do a get, get_extra
        m = r.get_memory(0)
        m = r.get_memory_extra(m)
        self.assertEqual(146520000, m.freq)
        # Now we should have the comment
        self.assertEqual('a comment', m.comment)

        # Make sure it is in the metadata
        self.assertIn('0000_comment', r.metadata['mem_extra'])

        # Erase the memory (only extra) and make sure we get no comment
        r.erase_memory_extra(0)

        # Make sure it's gone from metadata
        self.assertNotIn('0000_comment', r.metadata['mem_extra'])

        # Do a get, get_extra
        m = r.get_memory(0)
        m = r.get_memory_extra(m)
        self.assertEqual(146520000, m.freq)
        # Now we should have no comment because we erased
        self.assertEqual('', m.comment)

        r.set_memory_extra(m)

        # Make sure we don't keep empty comments
        self.assertNotIn('0000_comment', r.metadata['mem_extra'])


class TestOverrideRules(base.BaseTest):
    # You should not need to add your radio to this list. If you think you do,
    # please ask permission first.
    # Immutable fields should really only be used for cases where the value
    # is not changeable based on the *location* of the memory. If something
    # is forced to be a value based on the *content* of the memory (i.e. AM
    # for frequencies in airband), coerce them on set/get, and return a
    # ValidationWarning in validate_memory() so the user is told that the
    # values are being forced.
    IMMUTABLE_WHITELIST = [
        # Uncomment me when the time comes
        'Baofeng_GT-5R',
        'BTECH_GMRS-20V2',
        'BTECH_GMRS-50V2',
        'BTECH_GMRS-50X1',
        'BTECH_GMRS-V2',
        'BTECH_MURS-V2',
        'Radioddity_DB25-G',
        'Retevis_RB17P',
    ]

    def _test_radio_override_immutable_policy(self, rclass):
        self.assertEqual(
            chirp_common.Radio.check_set_memory_immutable_policy,
            rclass.check_set_memory_immutable_policy,
            'Radio %s should not override '
            'check_set_memory_immutable_policy' % (
                directory.radio_class_id(rclass)))

    def _test_radio_override_calls_super(self, rclass):
        r = rclass(None)
        method = r.check_set_memory_immutable_policy

        # Make sure the radio actually overrides it
        self.assertNotEqual(
            chirp_common.Radio.check_set_memory_immutable_policy,
            method,
            ('Radio %s in whitelist does not override '
             'check_set_memory_immutable_policy') % (
                 directory.radio_class_id(rclass)))

        # Make sure super() is called in the child class
        self.assertIn('__class__', method.__code__.co_freevars,
                      '%s.%s must call super() but does not' % (
                          rclass.__name__, method.__name__))

    def test_radio_overrides(self):
        for rclass in directory.DRV_TO_RADIO.values():
            if directory.radio_class_id(rclass) in self.IMMUTABLE_WHITELIST:
                self._test_radio_override_calls_super(rclass)
            else:
                self._test_radio_override_immutable_policy(rclass)


class TestMemory(base.BaseTest):
    def test_pickle_with_extra(self):
        m = chirp_common.Memory()
        m.extra = settings.RadioSettingGroup('extra', 'extra')
        m.extra.append(settings.RadioSetting(
            'test', 'test',
            settings.RadioSettingValueString(1, 32, current='foo')))
        n = pickle.loads(pickle.dumps(m))
        self.assertEqual(str(n.extra['test'].value),
                         str(m.extra['test'].value))

    def test_frozen_from_frozen(self):
        m = chirp_common.FrozenMemory(chirp_common.Memory(123))
        n = chirp_common.FrozenMemory(m)
        self.assertEqual(123, n.number)

    def test_frozen_dupe_unfrozen(self):
        FrozenMemory = chirp_common.FrozenMemory(
            chirp_common.Memory()).__class__
        m = chirp_common.FrozenMemory(chirp_common.Memory(123)).dupe()
        self.assertNotIsInstance(m, FrozenMemory)
        self.assertFalse(hasattr(m, '_frozen'))

    def test_frozen_modifications(self):
        orig = chirp_common.Memory(123)
        orig.extra = [settings.RadioSetting(
            'foo', 'Foo',
            settings.RadioSettingValueBoolean(False))]
        frozen = chirp_common.FrozenMemory(orig)
        with self.assertRaises(ValueError):
            frozen.extra[0].value = True

    def test_tone_validator(self):
        m = chirp_common.Memory()
        # 100.0 is a valid tone
        m.rtone = 100.0
        m.ctone = 100.0

        # 100 is not (must be a float)
        with self.assertRaises(ValueError):
            m.rtone = 100
        with self.assertRaises(ValueError):
            m.ctone = 100

        # 30.0 and 300.0 are out of range
        with self.assertRaises(ValueError):
            m.rtone = 30.0
        with self.assertRaises(ValueError):
            m.rtone = 300.0
        with self.assertRaises(ValueError):
            m.ctone = 30.0
        with self.assertRaises(ValueError):
            m.ctone = 300.0

    def test_repr_dump(self):
        m = chirp_common.Memory()
        self.assertEqual(
            "<Memory 0: freq=0,name='',vfo=0,rtone=88.5,ctone=88.5,dtcs=23,"
            "rx_dtcs=23,tmode='',cross_mode='Tone->Tone',dtcs_polarity='NN',"
            "skip='',power=None,duplex='',offset=600000,mode='FM',"
            "tuning_step=5.0,comment='',empty=False,immutable=[]>", repr(m))

        m.freq = 146520000
        m.rtone = 107.2
        m.tmode = 'Tone'
        self.assertEqual(
            "<Memory 0: freq=146520000,name='',vfo=0,rtone=107.2,ctone=88.5,"
            "dtcs=23,rx_dtcs=23,tmode='Tone',cross_mode='Tone->Tone',"
            "dtcs_polarity='NN',skip='',power=None,duplex='',offset=600000,"
            "mode='FM',tuning_step=5.0,comment='',empty=False,immutable=[]>",
            repr(m))

        m.number = 101
        m.extd_number = 'Call'
        self.assertEqual(
            "<Memory Call(101): freq=146520000,name='',vfo=0,rtone=107.2,"
            "ctone=88.5,dtcs=23,rx_dtcs=23,tmode='Tone',"
            "cross_mode='Tone->Tone',dtcs_polarity='NN',skip='',power=None,"
            "duplex='',offset=600000,mode='FM',tuning_step=5.0,comment='',"
            "empty=False,immutable=[]>", repr(m))

        m.extra = settings.RadioSettingGroup('extra', 'Extra')
        m.extra.append(
            settings.RadioSetting('test1', 'Test Setting 1',
                                  settings.RadioSettingValueBoolean(False)))
        m.extra.append(
            settings.RadioSetting('test2', 'Test Setting 2',
                                  settings.RadioSettingValueList(
                                      ['foo', 'bar'], 'foo')))
        self.assertEqual(
            "<Memory Call(101): freq=146520000,name='',vfo=0,rtone=107.2,"
            "ctone=88.5,dtcs=23,rx_dtcs=23,tmode='Tone',"
            "cross_mode='Tone->Tone',dtcs_polarity='NN',skip='',power=None,"
            "duplex='',offset=600000,mode='FM',tuning_step=5.0,comment='',"
            "empty=False,immutable=[],extra.test1='False',extra.test2='foo'>",
            repr(m))

    def test_debug_diff(self):
        m1 = chirp_common.Memory(1)
        m2 = chirp_common.Memory(1)

        m1.freq = 146520000
        m2.freq = 446000000
        self.assertEqual('freq=146520000>446000000', m1.debug_diff(m2, '>'))

        m2.tmode = 'TSQL'
        self.assertEqual("freq=146520000/446000000,tmode=''/'TSQL'",
                         m1.debug_diff(m2))

        # Make sure ident diffs come first and are noticed
        m2.number = 2
        m2.freq = 146520000
        self.assertEqual("ident=1/2,tmode=''/'TSQL'", m1.debug_diff(m2))

        # Make sure we can diff extras, and amongst heterogeneous formats
        m2.number = 1
        m1.extra = settings.RadioSettingGroup('extra', 'Extra')
        m2.extra = settings.RadioSettingGroup('extra', 'Extra')
        m1.extra.append(
            settings.RadioSetting('test1', 'Test Setting 1',
                                  settings.RadioSettingValueBoolean(False)))
        m2.extra.append(
            settings.RadioSetting('test2', 'Test Setting 2',
                                  settings.RadioSettingValueList(
                                      ['foo', 'bar'], 'foo')))
        self.assertEqual(
            "extra.test1='False'/'<missing>',extra.test2='<missing>'/'foo',"
            "tmode=''/'TSQL'", m1.debug_diff(m2))


class TestRadioFeatures(base.BaseTest):
    def test_valid_tones(self):
        rf = chirp_common.RadioFeatures()
        # These are valid tones
        rf.valid_tones = [100.0, 107.2]

        # These contain invalid tones
        with self.assertRaises(ValueError):
            rf.valid_tones = [100.0, 30.0]
        with self.assertRaises(ValueError):
            rf.valid_tones = [100.0, 300.0]
        with self.assertRaises(ValueError):
            rf.valid_tones = [100, 107.2]



import builtins
import unittest

import warnings
warnings.simplefilter('ignore', Warning)

builtins._ = lambda x: x


class BaseTest(unittest.TestCase):
    def setUp(self):
        __builtins__['_'] = lambda s: s
        self.mocks = []

    def use(self, m):
        self.mocks.append(m)
        m.start()

    def tearDown(self):
        for m in self.mocks:
            m.stop()



import logging
import sys
from unittest import mock

import pytest

from chirp import directory
from chirp import logger
from tests import base


class TestCaseDetect(base.DriverTest):
    def test_detect(self):
        with logger.log_history(logging.WARNING, 'chirp.drivers') as history:
            radio = directory.get_radio_by_image(self.TEST_IMAGE)
            self.assertEqual([], history.get_history(),
                             'Drivers should not log warnings/errors for '
                             'good images')
        if hasattr(radio, '_orig_rclass'):
            radio = radio._orig_rclass(self.TEST_IMAGE)
        if isinstance(self.radio, radio.__class__):
            # If we are a sub-device of the detected class then that's fine.
            # There's no good way for us to know that other than checking, so
            # the below assertion is just for show.
            self.assertIsInstance(self.radio, radio.__class__)
        else:
            self.assertIsInstance(radio, self.RADIO_CLASS,
                                  "Image %s detected as %s but expected %s" % (
                                      self.TEST_IMAGE,
                                      radio.__class__, self.RADIO_CLASS))

    @pytest.mark.skipif(sys.version_info < (3, 10),
                        reason="requires python3.10 or higher")
    @mock.patch('builtins.print')
    def test_match_model_is_quiet_no_match(self, mock_print):
        with self.assertNoLogs(level=logging.DEBUG):
            self.radio.match_model(b'', 'foo.img')
        mock_print.assert_not_called()

    @pytest.mark.skipif(sys.version_info < (3, 10),
                        reason="requires python3.10 or higher")
    @mock.patch('builtins.print')
    def test_match_model_is_quiet_with_match(self, mock_print):
        with self.assertNoLogs(level=logging.DEBUG):
            with open(self.TEST_IMAGE, 'rb') as f:
                self.radio.match_model(f.read(), self.TEST_IMAGE)
        mock_print.assert_not_called()



import datetime
import json
import os
import shutil
import tempfile
import unittest
from unittest import mock

import pytest

from chirp import chirp_common
from chirp.sources import repeaterbook


class TestRepeaterbook(unittest.TestCase):
    def fake_config(self, fn):
        return os.path.join(self.tempdir, fn)

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.patches = []
        self.patches.append(mock.patch('chirp.platform.Platform.config_file',
                            self.fake_config))
        for p in self.patches:
            p.start()

        self.testfile = os.path.join(os.path.dirname(__file__),
                                     'rb-united_states-oregon.json')

    def tearDown(self):
        for p in self.patches:
            p.stop()
        shutil.rmtree(self.tempdir)

    @pytest.mark.network
    def _test_get_state(self, state, coords=(None, None)):
        rb = repeaterbook.RepeaterBook()
        self.assertRaises(IndexError, rb.get_memory, 0)
        rb.do_fetch(mock.MagicMock(), {
            'country': 'United States',
            'state': state,
            'lat': coords[0],
            'lon': coords[1],
            'dist': 100,
            'openonly': False,
            'cached': False,
        })
        m = rb.get_memory(0)
        self.assertIsInstance(m, chirp_common.Memory)
        f = rb.get_features()
        self.assertGreater(sum(f.memory_bounds), 20)

        for i in range(*f.memory_bounds):
            m = rb.get_memory(i)
            if m.mode == 'DV':
                self.assertIsInstance(m, chirp_common.DVMemory)
                self.assertEqual('CQCQCQ  ', m.dv_urcall)
                self.assertEqual(8, len(m.dv_rpt1call))
                self.assertEqual(8, len(m.dv_rpt2call))
                self.assertEqual(m.dv_rpt1call, m.dv_rpt2call)
                self.assertNotEqual('', m.dv_rpt1call)
                break
        else:
            raise Exception('Did not find any DV results')

    @pytest.mark.network
    def test_get_oregon(self):
        self._test_get_state('Oregon', (45, -122))

    @pytest.mark.network
    def test_get_california(self):
        self._test_get_state('California')

    def _test_get_gmrs(self, state, coords=(None, None)):
        rb = repeaterbook.RepeaterBook()
        self.assertRaises(IndexError, rb.get_memory, 0)
        rb.do_fetch(mock.MagicMock(), {
            'country': 'United States',
            'state': state,
            'lat': coords[0],
            'lon': coords[1],
            'dist': 100,
            'service': 'gmrs',
            'openonly': False,
            'cached': False,
        })
        m = rb.get_memory(0)
        self.assertIsInstance(m, chirp_common.Memory)
        f = rb.get_features()
        self.assertGreater(sum(f.memory_bounds), 1)

    @pytest.mark.network
    def test_get_california_gmrs(self):
        self._test_get_gmrs('California')

    @pytest.mark.network
    def test_get_australia(self):
        rb = repeaterbook.RepeaterBook()
        self.assertRaises(IndexError, rb.get_memory, 0)
        rb.do_fetch(mock.MagicMock(), {
            'country': 'Australia',
            'state': 'ALL',
            'lat': -26,
            'lon': 133,
            'dist': 20000,
            'openonly': False,
            'cached': False,
        })
        m = rb.get_memory(0)
        self.assertIsInstance(m, chirp_common.Memory)
        f = rb.get_features()
        self.assertGreater(sum(f.memory_bounds), 20)

    def _test_with_mocked(self, params):
        rb = repeaterbook.RepeaterBook()
        with mock.patch.object(rb, 'get_data') as gd:
            gd.return_value = self.testfile
            rb.do_fetch(mock.MagicMock(), params)
        return rb

    def test_filter(self):
        params = {'country': 'United States',
                  'state': 'Oregon',
                  'lat': 45,
                  'lon': -122,
                  'dist': 0,
                  'openonly': False,
                  'cached': False,
                  'filter': 'tower'}
        rb = self._test_with_mocked(params)
        self.assertGreater(sum(rb.get_features().memory_bounds), 5)

    def test_distance(self):
        params = {'country': 'United States',
                  'state': 'Oregon',
                  'lat': 45,
                  'lon': -122,
                  'dist': 0,
                  'openonly': False,
                  'cached': False,
                  'filter': 'tower'}
        rb1 = self._test_with_mocked(dict(params))
        self.assertGreater(sum(rb1.get_features().memory_bounds), 5)

        # Make sure we got fewer than everything in Oregon
        params['dist'] = 100
        rb2 = self._test_with_mocked(dict(params))
        self.assertLess(sum(rb2.get_features().memory_bounds),
                        sum(rb1.get_features().memory_bounds))

        # There is nothing 1km away from this spot in the ocean
        params['dist'] = 1
        params['lat'] = 45
        params['lon'] = -140
        rb3 = self._test_with_mocked(dict(params))
        self.assertEqual(sum(rb3.get_features().memory_bounds), -1)

    def test_get_data_500(self):
        rb = repeaterbook.RepeaterBook()
        with mock.patch('requests.get') as mock_get:
            mock_get.return_value.status_code = 500
            status = mock.MagicMock()
            r = rb.get_data(status, 'US', 'OR', '')
            self.assertIsNone(r)
            status.send_fail.assert_called()
            # Make sure we wrote no files
            self.assertEqual([], os.listdir(os.path.join(self.tempdir,
                                                         'repeaterbook')))

    def test_get_data_json_fail(self):
        rb = repeaterbook.RepeaterBook()
        with mock.patch('requests.get') as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.iter_content.return_value = [b'foo']
            status = mock.MagicMock()
            r = rb.get_data(status, 'US', 'OR', '')
            self.assertIsNone(r)
            files = os.listdir(os.path.join(self.tempdir,
                                            'repeaterbook'))
            # Make sure we only wrote one file and that it is a tempfile
            # not one we will find as a data file later
            self.assertEqual(1, len(files))
            self.assertTrue(files[0].endswith('tmp'))
            status.send_fail.assert_called()

    def test_get_data_no_results(self):
        rb = repeaterbook.RepeaterBook()
        with mock.patch('requests.get') as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.iter_content.return_value = [json.dumps(
                {'count': 0}).encode()]
            status = mock.MagicMock()
            r = rb.get_data(status, 'US', 'OR', '')
            self.assertIsNone(r)
            files = os.listdir(os.path.join(self.tempdir,
                                            'repeaterbook'))
            # Make sure we cleaned up and signaled failure
            self.assertEqual(0, len(files))
            status.send_fail.assert_called()

    def test_get_data_got_results(self):
        files = os.listdir(self.tempdir)
        # Make sure we started with no data files
        self.assertEqual(0, len(files))
        rb = repeaterbook.RepeaterBook()
        with mock.patch('requests.get') as mock_get:
            mock_get.return_value.status_code = 200
            mock_get.return_value.iter_content.return_value = [json.dumps(
                {'count': 1}).encode()]
            status = mock.MagicMock()
            r = rb.get_data(status, 'US', 'OR', '')
            self.assertIsNotNone(r)
            files = os.listdir(os.path.join(self.tempdir,
                                            'repeaterbook'))
            # Make sure we left only a data file and did not signal failure
            self.assertEqual(1, len(files))
            self.assertFalse(files[0].endswith('tmp'))
            self.assertEqual(os.path.basename(r), files[0])
            status.send_fail.assert_not_called()

    def test_get_data_honors_cache_rules(self):
        os.mkdir(os.path.join(self.tempdir, 'repeaterbook'))
        cache_file = os.path.join(self.tempdir,
                                  'repeaterbook',
                                  os.path.basename(self.testfile))
        with open(cache_file, 'w') as f:
            f.write('foo')
        rb = repeaterbook.RepeaterBook()
        with mock.patch('requests.get') as mock_get:
            r = rb.get_data(mock.MagicMock(), 'United States', 'Oregon', '')
            self.assertEqual(cache_file, r)
            # Make sure we returned the cached file
            mock_get.assert_not_called()

            real_timedelta = datetime.timedelta
            real_datetime = datetime.datetime

            future = datetime.datetime.now() + datetime.timedelta(days=4)
            with mock.patch.object(repeaterbook, 'datetime') as mock_dt:
                mock_dt.datetime.fromtimestamp = real_datetime.fromtimestamp
                mock_dt.timedelta = real_timedelta
                mock_dt.datetime.now.return_value = future
                r = rb.get_data(mock.MagicMock(),
                                'United States', 'Oregon', '')
                # Cache file is 4 days old, we should use it
                mock_get.assert_not_called()

            future = datetime.datetime.now() + datetime.timedelta(days=45)
            with mock.patch.object(repeaterbook, 'datetime') as mock_dt:
                mock_dt.datetime.fromtimestamp = real_datetime.fromtimestamp
                mock_dt.timedelta = real_timedelta
                mock_dt.datetime.now.return_value = future
                r = rb.get_data(mock.MagicMock(),
                                'United States', 'Oregon', '')
                # Cache file is 45 days old, we should re-fetch
                mock_get.assert_called()



# Copyright 2024 Dan Smith <chirp@f.danplanet.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import datetime
import gzip
import logging
import os
import platform
import subprocess
import tempfile
import threading
import time

import requests
import wx
import wx.adv

from chirp import CHIRP_VERSION
from chirp import chirp_common
from chirp import logger
from chirp import platform as chirp_platform
from chirp.wxui import common
from chirp.wxui import config
from chirp.wxui import serialtrace

_ = wx.GetTranslation
CONF = config.get()
BASE = CONF.get('baseurl', 'chirpmyradio') or 'https://www.chirpmyradio.com'
LOG = logging.getLogger(__name__)
ReportThreadEvent, EVT_REPORT_THREAD = wx.lib.newevent.NewCommandEvent()


def get_chirp_platform():
    """Get the proper platform name for the chirp site's field"""
    p = platform.system()
    return 'MacOS' if p == 'Darwin' else p


def get_macos_system_info(manifest):
    try:
        sp = subprocess.check_output(
            'system_profiler SPSoftwareDataType SPUSBDataType',
            shell=True)
    except Exception as e:
        sp = 'Error getting system_profiler data: %s' % e
    manifest['files']['macos_system_info.txt'] = sp


def get_linux_system_info(manifest):
    try:
        sp = subprocess.check_output('lsusb',
                                     shell=True)
    except Exception as e:
        sp = 'Error getting system data: %s' % e
    manifest['files']['linux_system_info.txt'] = sp


def get_windows_system_info(manifest):
    try:
        sp = subprocess.check_output('pnputil /enum-devices /connected',
                                     shell=True)
    except Exception as e:
        sp = 'Error getting system data: %s' % e
    manifest['files']['win_system_info.txt'] = sp


@common.error_proof()
def prepare_report(chirpmain):
    manifest = {'files': {}}

    # Copy and clean/redact config
    config._CONFIG.save()
    conf_fn = chirp_platform.get_platform().config_file('chirp.config')
    LOG.debug('Capturing config file %s stamped %s', conf_fn,
              datetime.datetime.fromtimestamp(
                  os.stat(conf_fn).st_mtime).isoformat())
    with open(conf_fn, 'rb') as f:
        config_lines = f.readlines()
    clean_lines = []
    for line in list(config_lines):
        if b'password' in line:
            key, value = line.split(b'=', 1)
            value = b'***REDACTED***'
            line = b'%s = %s' % (key.strip(), value)
        clean_lines.append(line.strip())
    manifest['files']['config.txt'] = b'\n'.join(clean_lines)

    # Attach the currently-open file
    editor = chirpmain.current_editorset
    if editor and isinstance(editor._radio, chirp_common.FileBackedRadio):
        tmpf = tempfile.mktemp('-capture.img', 'chirp')
        LOG.debug('Capturing focused open file %s from %s',
                  editor.filename, editor._radio)
        editor._radio.save(tmpf)
        with open(tmpf, 'rb') as f:
            manifest['files'][os.path.basename(editor.filename)] = f.read()

    # Gather system details, if available
    system = platform.system()
    if system == 'Darwin':
        LOG.debug('Capturing macOS system_profiler data')
        get_macos_system_info(manifest)
    elif system == 'Linux':
        LOG.debug('Capturing linux system info')
        get_linux_system_info(manifest)
    elif system == 'Windows':
        LOG.debug('Capturing windows system info')
        get_windows_system_info(manifest)
    else:
        LOG.debug('No system info support for %s', system)

    # Snapshot debug log last
    if logger.Logger.instance.has_debug_log_file:
        tmp = common.temporary_debug_log()
        with open(tmp, 'rb') as f:
            manifest['files']['debug_log.txt'] = f.read()
        tmpf = tempfile.mktemp('.config', 'chirp')

    # Grab any trace files
    for tracefile in serialtrace.TRACEFILES:
        if os.path.exists(tracefile):
            LOG.debug('Capturing serial trace file %s', tracefile)
            with open(tracefile, 'rb') as f:
                manifest['files'][os.path.basename(tracefile)] = f.read()
        else:
            LOG.debug('Serial trace file %s does not exist', tracefile)

    return manifest


class BugReportContext:
    def __init__(self, wizard, chirpmain):
        self.wizard = wizard
        self.chirpmain = chirpmain
        self.editor = chirpmain.current_editorset
        self.session = requests.Session()
        self.session.headers = {
            'User-Agent': 'CHIRP/%s' % CHIRP_VERSION,
            'Referer': 'https://chirpmyradio.com/projects/chirp/issues/new',
        }

    def get_page(self, name, cls):
        if not hasattr(self, 'page_%s' % name):
            LOG.debug('Created page %s', name)
            setattr(self, 'page_%s' % name, cls(self))
        return getattr(self, 'page_%s' % name)


class BugReportPage(wx.adv.WizardPage):
    INST = ''

    def __init__(self, context):
        super().__init__(context.wizard)
        self.context = context
        vbox = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(vbox)

        inst = wx.StaticText(self, label=self.INST,
                             style=wx.TE_CENTER)
        inst.Wrap(self.wrap_width)
        vbox.Add(inst, 0, border=20, flag=wx.EXPAND | wx.ALL)

        self._build(vbox)

    @property
    def wrap_width(self):
        return 500

    def _validate_next(self):
        return True

    def validate_next(self, *a):
        self.FindWindowById(wx.ID_FORWARD).Enable(self._validate_next())

    def validate_success(self, event):
        pass


class Start(BugReportPage):
    INST = _(
        'This tool will upload details about your system to an existing '
        'issue on the CHIRP tracker. It requires your username and '
        'password for chirpmyradio.com in order to work. Information '
        'about your system, including your debug log, config file, and '
        'any open image files will be uploaded. An attempt will '
        'be made to redact any personal information before it leaves '
        'your system.'
        )

    def _build(self, vbox):
        self.context.is_new = None
        choices = wx.Panel(self)
        cvbox = wx.BoxSizer(wx.VERTICAL)
        choices.SetSizer(cvbox)

        self.newbug = wx.RadioButton(choices, label=_('File a new bug'),
                                     style=wx.RB_GROUP)
        self.newbug.Bind(wx.EVT_RADIOBUTTON, self.validate_next)
        cvbox.Add(self.newbug, flag=wx.ALIGN_LEFT)

        self.existbug = wx.RadioButton(choices,
                                       label=_('Update an existing bug'))
        self.existbug.Bind(wx.EVT_RADIOBUTTON, self.validate_next)
        cvbox.Add(self.existbug, flag=wx.ALIGN_LEFT)

        vbox.Add(choices, 1, flag=wx.ALIGN_CENTER)

        self.validate_next()

        if not logger.Logger.instance.has_debug_log_file:
            wx.MessageDialog(
                self,
                _('The debug log file is not available when CHIRP is run '
                  'interactively from the command-line. Thus, this tool will '
                  'not upload what you expect. It is recommended that you '
                  'quit now and run CHIRP non-interactively (or with stdin '
                  'redirected to /dev/null)'),
                _('Warning'),
                style=wx.OK | wx.ICON_WARNING).ShowModal()

    def GetNext(self):
        return self.context.get_page('creds', GetCreds)

    def _validate_next(self):
        return self.newbug.GetValue() or self.existbug.GetValue()

    def validate_success(self, event):
        self.context.is_new = self.newbug.GetValue()


class GetCreds(BugReportPage):
    INST = _('Enter your username and password for chirpmyradio.com. '
             'If you do not have an account click below to create '
             'one before proceeding.')

    def _build(self, vbox):
        vbox.Add(
            wx.adv.HyperlinkCtrl(
                self, label=BASE + '/account/register',
                url=BASE + '/account/register'),
            0, border=20, flag=wx.ALIGN_CENTER)

        panel = wx.Panel(self)
        vbox.Add(panel, 1, border=20, flag=wx.EXPAND | wx.ALL)
        grid = wx.FlexGridSizer(2, 5, 0)
        grid.AddGrowableCol(1)

        grid.Add(wx.StaticText(
            panel, label=_('Username')),
            border=20, flag=wx.ALIGN_CENTER | wx.RIGHT | wx.LEFT)
        self.username = wx.TextCtrl(
            panel, value=CONF.get('chirp_user',
                                  'chirpmyradio') or '')
        self.username.Bind(wx.EVT_TEXT, self.validate_next)
        grid.Add(self.username, 1, border=20,
                 flag=wx.EXPAND | wx.RIGHT | wx.LEFT)

        grid.Add(wx.StaticText(
            panel, label=_('Password')),
            border=20, flag=wx.EXPAND | wx.RIGHT | wx.LEFT)
        self.password = wx.TextCtrl(
            panel, style=wx.TE_PASSWORD,
            value=CONF.get_password('chirp_password',
                                    'chirpmyradio') or '')
        self.password.Bind(wx.EVT_TEXT, self.validate_next)
        grid.Add(self.password, 1, border=20,
                 flag=wx.EXPAND | wx.RIGHT | wx.LEFT)
        panel.SetSizer(grid)
        self.SetSizer(vbox)

        self.SetSize(480, 640)

    def _validate_next(self):
        return (len(self.username.GetValue()) > 2 and
                len(self.password.GetValue()) > 2)

    def _check_limit(self, uid):
        issue_limit = 3
        known_user = None

        try:
            start = datetime.datetime.now(datetime.timezone.utc)
            start.replace(microsecond=0)
            start -= datetime.timedelta(days=7)

            r = self.context.session.get(
                BASE + '/issues.json',
                params={'author_id': uid, 'limit': issue_limit,
                        'created_on': '>=%s' % start.date().isoformat()},
                auth=self.context.auth)
            open_issue_count = r.json()['total_count']

            # If they have open issues, check to see if they're known
            if open_issue_count > issue_limit:
                r = self.context.session.get(
                    BASE + '/users/%i.json' % uid,
                    params={'include': 'memberships'},
                    auth=self.context.auth)
                known_user = len(r.json()['user']['memberships']) > 0
        except Exception as e:
            LOG.exception('Failed to get known-user info: %s', e)
            return True

        if not known_user and open_issue_count > issue_limit:
            LOG.warning('User %i is not known and has %i open recent issues',
                        uid, open_issue_count)
            wx.MessageDialog(
                self,
                _('You have opened multiple issues within the last week. '
                  'CHIRP limits the number of issues you can open to avoid '
                  'abuse. If you really need to open another, please do so '
                  'via the website.'),
                style=wx.OK | wx.ICON_WARNING).ShowModal()
            return False
        return True

    def validate_success(self, event):
        self.context.auth = requests.auth.HTTPBasicAuth(
            self.username.GetValue(), self.password.GetValue())

        r = self.context.session.get(BASE + '/my/account.json',
                                     auth=self.context.auth)
        if r.status_code != 200:
            LOG.error('Login auth check failed: %s', r.reason)
            wx.MessageDialog(self,
                             _('Login failed: '
                               'Check your username and password'),
                             _('An error has occurred'),
                             style=wx.OK | wx.ICON_ERROR).ShowModal()
            event.Veto()
            return

        uid = r.json()['user']['id']
        LOG.debug('CHIRP login success as %i', uid)
        CONF.set('chirp_user', self.username.GetValue(), 'chirpmyradio')
        CONF.set_password('chirp_password', self.password.GetValue(),
                          'chirpmyradio')

        if self.context.is_new and not self._check_limit(uid):
            event.Veto()
            return

    def GetPrev(self):
        return self.context.page_start

    def GetNext(self):
        if self.context.is_new:
            return self.context.get_page('newbug', NewBugInfo)
        else:
            return self.context.get_page('existbug', ExistingBugInfo)


class NewBugInfo(BugReportPage):
    INST = _('Enter information about the bug including a short but '
             'meaningful subject and information about the radio model '
             '(if applicable) in question. In the next step you will have '
             'a chance to add more details about the problem.')

    def _build(self, vbox):
        self.context.bugsubj = self.context.bugmodel = None
        panel = wx.Panel(self)
        vbox.Add(panel, 1, border=20, flag=wx.EXPAND)
        grid = wx.FlexGridSizer(2, 5, 0)
        grid.AddGrowableCol(1)
        panel.SetSizer(grid)

        grid.Add(wx.StaticText(
            panel, label=_('Bug subject:')),
            border=20, flag=wx.ALIGN_CENTER | wx.RIGHT | wx.LEFT)
        self.subj = wx.TextCtrl(panel)
        self.subj.SetMaxLength(100)
        self.subj.Bind(wx.EVT_TEXT, self.validate_next)
        grid.Add(self.subj, 1, border=20,
                 flag=wx.EXPAND | wx.RIGHT | wx.LEFT)

        grid.Add(wx.StaticText(
            panel, label=_('Radio model:')),
            border=20, flag=wx.ALIGN_CENTER | wx.RIGHT | wx.LEFT)
        self.model = wx.TextCtrl(panel)
        grid.Add(self.model, 1, border=20,
                 flag=wx.EXPAND | wx.RIGHT | wx.LEFT)

        if self.context.editor and self.context.editor._radio:
            self.model.SetValue('%s %s %s' % (
                self.context.editor._radio.VARIANT,
                self.context.editor._radio.VENDOR,
                self.context.editor._radio.MODEL))

    def _validate_next(self):
        return len(self.subj.GetValue()) > 10

    def validate_success(self, *a):
        self.context.bugsubj = self.subj.GetValue()
        self.context.bugmodel = self.model.GetValue()

    def GetPrev(self):
        return self.context.page_creds

    def GetNext(self):
        return self.context.get_page('update', BugUpdateInfo)


class ExistingBugInfo(BugReportPage):
    INST = _('Enter the bug number that should be updated')

    def _build(self, vbox):
        self.context.bugnum = None
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        panel = wx.Panel(self)
        panel.SetMinSize((self.wrap_width, -1))
        panel.SetSizer(hbox)

        hbox.Add(wx.StaticText(panel, label=_('Bug number:')),
                 0, border=20, flag=wx.LEFT | wx.RIGHT)
        self.bugnum = wx.TextCtrl(panel)
        self.bugnum.Bind(wx.EVT_TEXT, self.validate_next)
        self.bugnum.SetToolTip(
            _('This is the ticket number for an already-created issue on the '
              'chirpmyradio.com website'))
        hbox.Add(self.bugnum, 1, border=20, flag=wx.EXPAND)

        vbox.Add(panel, 0, border=20, flag=wx.ALIGN_CENTER)

    def _validate_next(self):
        return self.bugnum.GetValue().isdigit()

    def validate_success(self, event):
        self.context.bugnum = self.bugnum.GetValue()
        r = self.context.session.get(
            BASE + '/issues/%s.json' % self.context.bugnum,
            auth=self.context.auth)
        if r.status_code != 200:
            LOG.error('Failed to access issue %s: %i %s',
                      self.context.bugnum, r.status_code, r.reason)
            wx.MessageDialog(self,
                             _('Bug number not found'),
                             _('An error has occurred'),
                             style=wx.OK | wx.ICON_ERROR).ShowModal()
            event.Veto()
        elif r.json()['issue']['status']['is_closed']:
            LOG.error('Issue %s is closed', self.context.bugnum)
            wx.MessageDialog(
                self,
                _('Please check the bug number and try again. If you do '
                  'actually want to report against this bug, please comment '
                  'on it via the website and ask for it to be re-opened.'),
                _('Bug number %s is closed') % self.context.bugnum
                ).ShowModal()
            event.Veto()
        else:
            LOG.debug('Validated issue %s', self.context.bugnum)

    def GetNext(self):
        return self.context.get_page('update', BugUpdateInfo)

    def GetPrev(self):
        return self.context.page_creds


class BugUpdateInfo(BugReportPage):
    INST = _('Enter details about this update. Be descriptive about what '
             'you were doing, what you expected to happen, and what '
             'actually happened.')
    DEFAULT = '\n'.join([
            _('(Describe what you were doing)'),
            '',
            _('(Describe what you expected to happen)'),
            '',
            _('(Describe what actually happened instead)'),
            '',
            _('(Has this ever worked before? New radio? '
                'Does it work with OEM software?)'),
    ])

    def _build(self, vbox):
        self.context.bugdetails = None
        self.details = wx.TextCtrl(self, style=wx.TE_MULTILINE,
                                   value=self.DEFAULT)
        try:
            self.details.SetHint(_('Enter information to add to the bug here'))
        except Exception:
            # Older wx doesn't allow this on multi-line fields (?)
            pass
        self.details.Bind(wx.EVT_TEXT, self.validate_next)
        vbox.Add(self.details, 1, border=20,
                 flag=wx.EXPAND | wx.LEFT | wx.RIGHT)

    def _validate_next(self):
        current = self.details.GetValue()
        return len(current) > 10 and current != self.DEFAULT

    def validate_success(self, *a):
        self.context.bugdetails = self.details.GetValue()

    def GetNext(self):
        # Always create a fresh one so it's updated
        return SubmitPage(self.context)

    def GetPrev(self):
        if self.context.is_new:
            return self.context.get_page('newbug', NewBugInfo)
        else:
            return self.context.get_page('existbug', ExistingBugInfo)


class SubmitPage(BugReportPage):
    INST = _('The following information will be submitted:')

    def _build(self, vbox):
        self.deets = wx.TextCtrl(self, style=wx.TE_MULTILINE | wx.TE_READONLY)
        vbox.Add(self.deets, 1, border=20, flag=wx.EXPAND)

    def update(self):
        self.context.manifest = prepare_report(self.context.chirpmain)
        self.context.manifest['desc'] = self.context.bugdetails
        if self.context.is_new:
            self.context.manifest['bugsubj'] = self.context.bugsubj
            self.context.manifest['bugmodel'] = self.context.bugmodel
        else:
            self.context.manifest['issue'] = self.context.bugnum
        text = '\n'.join([
            _('Reporting a new bug: %r') % self.context.bugsubj if
            self.context.is_new else
            _('Updating bug %s') % self.context.bugnum,
            _('Files:') + ' ' + ','.join(
                self.context.manifest['files'].keys()),
            _('Detailed information') + ':\n',
            self.context.bugdetails or '',
            ])
        self.deets.SetValue(text)

    def _validate_next(self):
        self.update()
        return True

    def GetPrev(self):
        return self.context.page_update

    def GetNext(self):
        return ResultPage(self.context)


class ResultPage(BugReportPage):
    def _build(self, vbox):
        self.ran = False
        self.thread = None
        self.Bind(EVT_REPORT_THREAD, self.report_done)

        self.result = wx.StaticText(self, label='Sending report...',
                                    style=wx.TE_CENTER)
        vbox.Add(self.result, 0, border=20, flag=wx.EXPAND | wx.ALL)

        self.issuelink = wx.adv.HyperlinkCtrl(self, url=BASE)
        self.issuelink.Hide()
        vbox.Add(self.issuelink, 0, border=20, flag=wx.EXPAND | wx.ALL)

    def start_thread(self):
        self.ran = True
        LOG.debug('Preparing to submit report')

        self.thread = threading.Thread(target=self.send_report,
                                       args=(self.context.manifest,))
        self.thread.start()

    def _validate_next(self):
        if not self.ran:
            self.start_thread()
        return self.thread is None

    def send_report(self, manifest):
        LOG.info('Report thread running for issue %s',
                 manifest.get('issue', '(new)'))
        try:
            result = self._send_report(manifest)
            failed = False
        except Exception as e:
            LOG.exception('Failed to report: %s', e)
            result = str(e)
            failed = True
        wx.PostEvent(self, ReportThreadEvent(self.GetId(),
                                             result=result, failed=failed))

    def _create_bug(self, manifest):
        issue_data = {
            'issue': {
                'project_id': 1,
                'priority_id': 4,
                'subject': manifest['bugsubj'],
                'description': manifest['desc'],
                'custom_fields': [
                    {'id': 1, 'value': 'next'},
                    {'id': 2, 'value': manifest['bugmodel']},
                    {'id': 3, 'value': get_chirp_platform()},
                    {'id': 7, 'value': '1'},  # read instructions
                ]
            }
        }
        r = self.context.session.post(BASE + '/issues.json',
                                      json=issue_data,
                                      auth=self.context.auth)
        if r.status_code != 201:
            LOG.error('Failed to create issue: %i %s', r.status_code, r.reason)
            raise Exception('Failed to create new issue')
        manifest['issue'] = str(r.json()['issue']['id'])
        self.context.bugnum = manifest['issue']
        LOG.info('Created new issue %s', manifest['issue'])

    def _upload_file(self, manifest, fn):
        for i in range(3):
            LOG.debug('Uploading %s attempt %i', fn, i + 1)
            try:
                r = self.context.session.post(
                    BASE + '/uploads.json',
                    params={'filename': fn},
                    data=manifest['files'][fn],
                    headers={
                        'Content-Type': 'application/octet-stream'},
                    auth=self.context.auth)
            except Exception as e:
                LOG.error('Exception uploading %s: %s', fn, e)
                time.sleep(2 + (2 * i))
                continue
            if r.status_code >= 400:
                LOG.error('Failed to upload %s: %s %s',
                          fn, r.status_code, r.reason)
                time.sleep(2 + (2 * i))
                continue
            elif r.status_code != 201:
                LOG.error('Failed to upload %s: %s %s',
                          fn, r.status_code, r.reason)
                raise Exception('Failed to upload file: %s' % r.reason)
            return r.json()['upload']['token']
        raise Exception('Failed to upload %s after multiple attempts', fn)

    def _send_report(self, manifest):
        if 'issue' not in manifest:
            self._create_bug(manifest)

        for fn in list(manifest['files'].keys()):
            fdata = manifest['files'][fn]
            if len(fdata) > 1024 * 1024:
                LOG.warning('File %s is larger than 1MB, compressing', fn)
                fdata = gzip.compress(fdata)
                manifest['files'].pop(fn)
                fn += '.gz'
                manifest['files'][fn] = fdata

        notes = '[Uploaded from CHIRP %s]\n\n' % CHIRP_VERSION
        tokens = []
        for fn in manifest['files']:
            try:
                token = self._upload_file(manifest, fn)
            except Exception as e:
                LOG.error('Failed to upload file %s: %s', fn, e)
                notes += '[Failed to upload file %s: %s]\n\n' % (fn, e)
                continue
            ext = os.path.splitext(fn)[1].lower()
            if ext in ('.log', '.txt'):
                ct = 'text/plain'
            else:
                ct = 'application/octet-stream'
            token_info = {'token': token,
                          'filename': fn,
                          'content_type': ct}
            tokens.append(token_info)

        LOG.debug('File tokens: %s', tokens)

        if not self.context.is_new:
            notes += manifest['desc']
        r = self.context.session.put(
            BASE + '/issues/%s.json' % manifest['issue'],
            json={'issue': {
                    'notes': notes,
                    'uploads': tokens,
                }},
            auth=self.context.auth)
        if r.status_code != 204:
            LOG.error('Failed to update issue %s with tokens %s: %s %s',
                      manifest['issue'], tokens, r.status_code, r.reason)
            raise Exception('Failed to update issue')

    def report_done(self, event):
        self.thread = None
        LOG.info('Report thread returned %s', event.result)
        if event.failed:
            self.result.SetLabel(
                _('Failed to send bug report:') + '\n' + event.result)
        else:
            self.result.SetLabel(
                _('Successfully sent bug report:'))
            self.FindWindowById(wx.ID_BACKWARD).Enable(False)
            link = BASE + '/issues/%s' % self.context.bugnum
            self.issuelink.SetLabel(link)
            self.issuelink.SetURL(link)
            self.issuelink.Show()
        self.validate_next()
        self.GetSizer().Layout()

    def GetNext(self):
        return None

    def GetPrev(self):
        return self.context.page_submit


def do_bugreport(parent, event):
    wizard = wx.adv.Wizard(parent)

    wizard.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGED,
                lambda e: e.GetPage().validate_next())
    wizard.Bind(wx.adv.EVT_WIZARD_PAGE_CHANGING,
                lambda e: (e.GetPage().validate_success(e)
                           if e.GetDirection() else None))

    wizard.SetPageSize((640, 400))
    context = BugReportContext(wizard, parent)
    start = context.get_page('start', Start)
    wizard.GetPageAreaSizer().Add(start)
    wizard.RunWizard(start)
    wizard.Destroy()



import logging

import pytest

from chirp import chirp_common
from chirp import errors
from tests import base

LOG = logging.getLogger(__name__)


class TestCaseBruteForce(base.DriverTest):
    def set_and_compare(self, m, **kwargs):
        msgs = self.radio.validate_memory(m)
        if msgs:
            # If the radio correctly refuses memories it can't
            # store, don't fail
            return

        self.radio.set_memory(chirp_common.FrozenMemory(m))
        ret_m = self.radio.get_memory(m.number)

        # Damned Baofeng radios don't seem to properly store
        # shift and direction, so be gracious here
        if m.duplex == "split" and ret_m.duplex in ["-", "+"]:
            ret_m.offset = ret_m.freq + \
                (ret_m.offset * int(ret_m.duplex + "1"))
            ret_m.duplex = "split"

        self.assertEqualMem(m, ret_m, **kwargs)

    def test_tone(self):
        m = self.get_mem()
        for tone in chirp_common.TONES:
            for tmode in self.rf.valid_tmodes:
                if tmode not in chirp_common.TONE_MODES:
                    continue
                elif tmode in ["DTCS", "DTCS-R", "Cross"]:
                    continue  # We'll test DCS and Cross tones separately

                m.tmode = tmode
                if tmode == "":
                    pass
                elif tmode == "Tone":
                    m.rtone = tone
                elif tmode in ["TSQL", "TSQL-R"]:
                    if self.rf.has_ctone:
                        m.ctone = tone
                    else:
                        m.rtone = tone
                else:
                    self.fail("Unknown tone mode `%s'" % tmode)

                try:
                    self.set_and_compare(m)
                except errors.UnsupportedToneError:
                    # If a radio doesn't support a particular tone value,
                    # don't punish it
                    pass

    @base.requires_feature('has_dtcs')
    def test_dtcs(self):
        m = self.get_mem()
        m.tmode = "DTCS"
        for code in self.rf.valid_dtcs_codes:
            m.dtcs = code
            self.set_and_compare(m)

        if not self.rf.has_dtcs_polarity:
            return

        for pol in self.rf.valid_dtcs_pols:
            m.dtcs_polarity = pol
            self.set_and_compare(m)

    @base.requires_feature('has_cross')
    def test_cross(self):
        m = self.get_mem()
        m.tmode = "Cross"
        # No fair asking a radio to detect two identical tones as Cross instead
        # of TSQL
        m.rtone = 100.0
        m.ctone = 107.2
        m.dtcs = 506
        m.rx_dtcs = 516
        for cross_mode in self.rf.valid_cross_modes:
            m.cross_mode = cross_mode
            self.set_and_compare(m)

    @base.requires_feature('valid_duplexes')
    def test_duplex(self):
        m = self.get_mem()
        if 'duplex' in m.immutable:
            self.skipTest('Test memory has immutable duplex')
        for duplex in self.rf.valid_duplexes:
            assert duplex in ["", "-", "+", "split", "off"]
            if duplex == 'split':
                self.assertTrue(self.rf.can_odd_split,
                                'Radio supports split but does not set '
                                'can_odd_split=True in features')
                m.offset = self.rf.valid_bands[0][1] - 100000
            else:
                m.offset = chirp_common.to_kHz(int(m.tuning_step) * 2)
            m.duplex = duplex
            # Ignore the offset because we do some fudging on this and we
            # don't necessarily know the best step to use. What we care about
            # is duplex here.
            self.set_and_compare(m, ignore=['offset'])

        if self.rf.can_odd_split:
            self.assertIn('split', self.rf.valid_duplexes,
                          'Radio claims can_odd_split but split not in '
                          'valid_duplexes')

    @base.requires_feature('valid_skips')
    def test_skip(self):
        mem = self.get_mem()
        lo, hi = self.rf.memory_bounds
        # Walk through several memories, specifically across 8 and 16,
        # as many radio use bitfields for skip flags.
        for i in range(max(5, lo), min(25, hi)):
            # Walk through the skip flags twice each, to make sure we can
            # toggle them on..off..on
            for skip in self.rf.valid_skips * 2:
                m = mem.dupe()
                if 'empty' not in m.immutable:
                    m.empty = False
                m.number = i
                m.skip = skip
                self.set_and_compare(m)
                # Delete the memory each time because some radios are
                # dynamically allocated and will run out of space here.
                self.radio.erase_memory(m.number)

    @base.requires_feature('valid_modes')
    def test_mode(self):
        m = self.get_mem()
        if 'mode' in m.immutable:
            self.skipTest('Test memory has immutable duplex')

        def ensure_urcall(call):
            lst = self.radio.get_urcall_list()
            lst[0] = call
            self.radio.set_urcall_list(lst)

        def ensure_rptcall(call):
            lst = self.radio.get_repeater_call_list()
            lst[0] = call
            self.radio.set_repeater_call_list(lst)

        def freq_is_ok(freq):
            for lo, hi in self.rf.valid_bands:
                if freq > lo and freq < hi:
                    return True
            return False

        successes = 0
        for mode in self.rf.valid_modes:
            self.assertIn(mode, chirp_common.MODES,
                          'Radio exposes non-standard mode')
            tmp = m.dupe()
            if mode == "DV" and \
                isinstance(self.radio,
                           chirp_common.IcomDstarSupport):
                tmp = chirp_common.DVMemory()
                try:
                    ensure_urcall(tmp.dv_urcall)
                    ensure_rptcall(tmp.dv_rpt1call)
                    ensure_rptcall(tmp.dv_rpt2call)
                except IndexError:
                    if self.rf.requires_call_lists:
                        raise
                    else:
                        # This radio may not do call lists at all,
                        # so let it slide
                        pass
            if mode == "FM" and freq_is_ok(tmp.freq + 100000000):
                # Some radios don't support FM below approximately 30MHz,
                # so jump up by 100MHz, if they support that
                tmp.freq += 100000000

            tmp.mode = mode

            if self.rf.validate_memory(tmp):
                # A result (of error messages) from validate means the radio
                # thinks this is invalid, so don't fail the test
                LOG.warning('Failed to validate %s: %s' % (
                    tmp, self.rf.validate_memory(tmp)))
                continue

            # Ignore tuning_step because changing modes may cause step changes
            # in some radios
            self.set_and_compare(tmp, ignore=['tuning_step'])
            successes += 1

    def test_validate_all(self):
        lo, hi = self.rf.memory_bounds
        for i in range(lo, hi + 1):
            m1 = self.radio.get_memory(i)
            if m1.empty:
                continue
            errs, warns = chirp_common.split_validation_msgs(
                self.radio.validate_memory(m1))
            self.assertEqual([], errs,
                             ('Radio has validation errors for memory %i '
                              'stored in sample image: %s') % (i, m1))

    @base.requires_feature('has_nostep_tuning', equal=False)
    def test_validate_all_steps(self):
        lo, hi = self.rf.memory_bounds
        for i in range(lo, hi + 1):
            m1 = self.radio.get_memory(i)
            try:
                chirp_common.required_step(m1.freq, self.rf.valid_tuning_steps)
            except Exception as e:
                self.fail('Memory %i: %s' % (i, e))

    @pytest.mark.skip(reason='not ready yet')
    def test_get_set_all(self):
        lo, hi = self.rf.memory_bounds
        for i in range(lo, hi + 1):
            m1 = self.radio.get_memory(i)
            self.radio.set_memory(m1)
            m2 = self.radio.get_memory(i)
            self.assertEqualMem(m1, m2)



import logging

from chirp import logger
from tests.unit import base


class TestLogger(base.BaseTest):
    def test_log_history(self):
        drv_log = logging.getLogger('chirp.drivers.foo')
        ui_log = logging.getLogger('chirp.wxui.foo')
        root_log = logging.getLogger()

        root_log.setLevel(logging.DEBUG)

        def log_all():
            for log in (drv_log, ui_log, root_log):
                log.debug('debug')
                log.warning('warning')
                log.error('error')

        # Log to everything
        log_all()

        with logger.log_history(logging.WARNING, 'chirp.drivers') as h:
            log_all()
            history = h.get_history()

        # We should only have the error,warning messages from drivers
        # since we started the capture, not before
        self.assertEqual(2, len(history))

        with logger.log_history(logging.WARNING, 'chirp.drivers') as h:
            log_all()
            history = h.get_history()

        # Make sure we only have the captured logs, not any leftover from
        # the previous run
        self.assertEqual(2, len(history))



# Copyright 2025 Jim Unroe <rock.unroe@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import logging
import struct

from chirp import (
    bitwise,
    chirp_common,
    directory,
    errors,
    memmap,
    util,
)
from chirp.drivers import h777
from chirp.settings import (
    MemSetting,
    RadioSettingGroup,
    RadioSettings,
    RadioSettingValueBoolean,
    RadioSettingValueInvertedBoolean,
    RadioSettingValueInteger,
    RadioSettingValueList,
)

LOG = logging.getLogger(__name__)

MEM_FORMAT = """
struct {
  ul32 rxfreq;        // 0-3
  ul32 txfreq;        // 4-7
  lbcd rx_tone[2];    // 8-9
  lbcd tx_tone[2];    // A-B
  u8 unknown0:2,      // C
     scramb:1,        //     Scramble  0 = Off, 1 = On
     scanadd:1,       //     Scan Add  0 = Scan, 1 = Skip
     ishighpower:1,   //     Power Level  0 = Low, 1 = High
     narrow:1,        //     Bandwidth  0 = Wide, 1 = Narrow
     unknown1:1,      //
     bcl:1;           //     Busy Channel Locklut  0 = On, 1 = Off
} memory[16];

struct {
  u8 codesw:1,        // 00D0  Code Switch
     scanm:1,         //       Scan Mode
     voxs:1,          //       VOX Switch
     roger:1,         //       Roger
     voice:1,         //       Voice Annunciation
     unknown_0:1,
     save:1,          //       Battery Save
     beep:1;          //       Beep Tone
  u8 squelch;         // 00D1  Squelch Level
  u8 tot;             // 00D2  Time-out Timer
  u8 vox;             // 00D3  VOX Level
  u8 voxd;            // 00D4  Vox Delay
  u8 unknown_1;
  u8 unknown_2;
  u8 skey2;           // 00D7  Side Key 2 (long)
  u8 unknown_3[5];    // 00D8 - 00DC
  u8 password[6];     // 00DD - 00E2  Password
} settings;
"""


CMD_ACK = b"\x06"

DTCS = tuple(sorted(chirp_common.DTCS_CODES + (645,)))

OFF1TO9_LIST = ["Off"] + ["%s" % x for x in range(1, 10)]
SCANM_LIST = ["Carrier", "Time"]
SKEY2_LIST = ["Off", "VOX", "Power", "Scan"]
TIMEOUTTIMER_LIST = ["Off"] + ["%s seconds" % x for x in range(30, 210, 30)]
VOICE_LIST = ["Off", "English"]
VOXD_LIST = ["0.5", "1.0", "1.5", "2.0", "2.5", "3.0"]


def _read_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'R', block_addr, block_size)
    expectedresponse = b"W" + cmd[1:]
    LOG.debug("Reading block %04x..." % (block_addr))

    try:
        serial.write(cmd)
        response = serial.read(4 + block_size)
        if response[:4] != expectedresponse:
            raise Exception("Error reading block %04x." % (block_addr))

        block_data = response[4:]
    except Exception:
        raise errors.RadioError("Failed to read block at %04x" % block_addr)

    return block_data


def _write_block(radio, block_addr, block_size):
    serial = radio.pipe

    cmd = struct.pack(">cHb", b'W', block_addr, block_size)
    data = radio.get_mmap()[block_addr:block_addr + block_size]

    LOG.debug("Writing Data:")
    LOG.debug(util.hexprint(cmd + data))

    try:
        serial.write(cmd + data)
        if serial.read(1) != CMD_ACK:
            raise Exception("No ACK")
    except Exception:
        raise errors.RadioError("Failed to send block "
                                "to radio at %04x" % block_addr)


def do_download(radio):
    LOG.debug("download")

    data = b""

    status = chirp_common.Status()
    status.msg = "Cloning from radio"

    status.cur = 0
    status.max = radio._memsize

    for addr in range(0, radio._memsize, radio.BLOCK_SIZE):
        status.cur = addr + radio.BLOCK_SIZE
        radio.status_fn(status)

        block = _read_block(radio, addr, radio.BLOCK_SIZE)
        data += block

        LOG.debug("Address: %04x" % addr)
        LOG.debug(util.hexprint(block))

    return memmap.MemoryMapBytes(data)


def do_upload(radio):
    status = chirp_common.Status()
    status.msg = "Uploading to radio"

    h777._h777_enter_single_programming_mode(radio)

    status.cur = 0
    status.max = radio._memsize

    for start_addr, end_addr in radio._ranges:
        for addr in range(start_addr, end_addr, radio.BLOCK_SIZE_UP):
            status.cur = addr + radio.BLOCK_SIZE_UP
            radio.status_fn(status)
            _write_block(radio, addr, radio.BLOCK_SIZE_UP)


class H777V4BaseRadio(chirp_common.CloneModeRadio):
    """RETEVIS H777 V4 Base"""
    VENDOR = "Retevis"
    MODEL = "H777 V4 Base"
    BAUD_RATE = 9600
    BLOCK_SIZE = 0x0D
    BLOCK_SIZE_UP = 0x0D

    POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=0.50),
                    chirp_common.PowerLevel("High", watts=2.00)
                    ]

    PROGRAM_CMD = b"C777HAM"
    _ranges = [
               (0x0000, 0x00EA),
              ]
    _memsize = 0x00EA

    def get_features(self):
        rf = chirp_common.RadioFeatures()
        rf.has_settings = True
        rf.has_bank = False
        rf.has_ctone = True
        rf.has_cross = True
        rf.has_rx_dtcs = True
        rf.has_tuning_step = False
        rf.can_odd_split = True
        rf.has_name = False
        rf.valid_skips = ["", "S"]
        rf.valid_tmodes = ["", "Tone", "TSQL", "DTCS", "Cross"]
        rf.valid_cross_modes = ["Tone->Tone", "Tone->DTCS", "DTCS->Tone",
                                "->Tone", "->DTCS", "DTCS->", "DTCS->DTCS"]
        rf.valid_power_levels = self.POWER_LEVELS
        rf.valid_duplexes = ["", "-", "+", "split", "off"]
        rf.valid_modes = ["FM", "NFM"]  # 25 kHz, 12.5 kHz.
        rf.valid_dtcs_codes = DTCS
        rf.memory_bounds = (1, 16)
        rf.valid_tuning_steps = [2.5, 5., 6.25, 10., 12.5, 20., 25., 50.]
        rf.valid_bands = [(400000000, 520000000)]
        return rf

    def process_mmap(self):
        self._memobj = bitwise.parse(MEM_FORMAT, self._mmap)

    def sync_in(self):
        """Download from radio"""
        try:
            data = do_download(self)
        except errors.RadioError:
            # Pass through any real errors we raise
            raise
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during download')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')
        self._mmap = data
        self.process_mmap()

    def sync_out(self):
        """Upload to radio"""
        try:
            do_upload(self)
        except errors.RadioError:
            raise
        except Exception:
            # If anything unexpected happens, make sure we raise
            # a RadioError and log the problem
            LOG.exception('Unexpected error during upload')
            raise errors.RadioError('Unexpected error communicating '
                                    'with the radio')

    def _decode_tone(self, toneval, txrx, chnum):
        pol = "N"
        rawval = (toneval[1].get_bits(0xFF) << 8) | toneval[0].get_bits(0xFF)

        if toneval[0].get_bits(0xFF) == 0xFF:
            mode = ""
            val = 0
        elif toneval[1].get_bits(0xC0) == 0xC0:
            val = int(f'{rawval & 0x1FF:o}')
            if val in DTCS:
                mode = "DTCS"
                pol = "R"
            else:
                LOG.error('unknown value: CH# %i DTCS %fR %s' % (
                    chnum, val, txrx))
                mode = ""
                val = 0
        elif toneval[1].get_bits(0x80):
            val = int(f'{rawval & 0x1FF:o}')
            if val in DTCS:
                mode = "DTCS"
            else:
                LOG.error('unknown value: CH# %i DTCS %fN %s' % (
                    chnum, val, txrx))
                mode = ""
                val = 0
        else:
            val = int(toneval) / 10.0
            if val in chirp_common.TONES:
                mode = "Tone"
            else:
                LOG.error('unknown value: CH# %i CTCSS %fHZ %s' % (
                    chnum, val, txrx))
                mode = ""
                val = 0

        return mode, val, pol

    def _encode_tone(self, _toneval, mode, val, pol):
        toneval = 0
        if mode == "Tone":
            toneval = int("%i" % (val * 10), 16)
        elif mode == "DTCS":
            toneval = int('%i' % val, 8)
            toneval |= 0x8000
            if pol == "R":
                toneval |= 0x4000
        else:
            toneval = 0xFFFF

        _toneval[0].set_raw(toneval & 0xFF)
        _toneval[1].set_raw((toneval >> 8) & 0xFF)

    def get_raw_memory(self, number):
        return repr(self._memobj.memory[number - 1])

    def get_memory(self, number):
        _mem = self._memobj.memory[number - 1]

        mem = chirp_common.Memory()
        mem.number = number

        if _mem.get_raw()[:1] == b"\xFF":
            mem.empty = True
            return mem

        mem.freq = int(_mem.rxfreq) * 10

        if _mem.txfreq == 0xFFFFFFFF:
            # TX freq not set
            mem.duplex = "off"
            mem.offset = 0
        elif abs(int(_mem.rxfreq) * 10 - int(_mem.txfreq) * 10) > 25000000:
            mem.duplex = "split"
            mem.offset = int(_mem.txfreq) * 10
        elif int(_mem.rxfreq) == int(_mem.txfreq):
            mem.duplex = ""
            mem.offset = 0
        else:
            mem.duplex = int(_mem.rxfreq) > int(_mem.txfreq) \
                and "-" or "+"
            mem.offset = abs(int(_mem.rxfreq) - int(_mem.txfreq)) * 10

        txmode, txval, txpol = self._decode_tone(_mem.tx_tone, 'TX', number)
        rxmode, rxval, rxpol = self._decode_tone(_mem.rx_tone, 'RX', number)

        chirp_common.split_tone_decode(mem,
                                       (txmode, txval, txpol),
                                       (rxmode, rxval, rxpol))

        if _mem.scanadd:
            mem.skip = "S"

        mem.power = self.POWER_LEVELS[_mem.ishighpower]

        mem.mode = _mem.narrow and "NFM" or "FM"

        mem.extra = RadioSettingGroup("Extra", "extra")

        # BCL (Busy Channel Lockout)
        rs = RadioSettingValueInvertedBoolean(not bool(_mem.bcl))
        rset = MemSetting("bcl", "BCL", rs)
        rset.set_doc("Busy Channel Lockout")
        mem.extra.append(rset)

        # Scramble
        rs = RadioSettingValueInvertedBoolean(not bool(_mem.scramb))
        rset = MemSetting("scramb", "Scramble", rs)
        rset.set_doc("Frequency inversion Scramble")
        mem.extra.append(rset)

        return mem

    def set_memory(self, mem):
        _mem = self._memobj.memory[mem.number - 1]

        _mem.set_raw(b"\xff" * 13)
        if mem.empty:
            return

        _mem.rxfreq = mem.freq / 10
        if mem.duplex == "off":
            _mem.txfreq = 0xFFFFFFFF
        elif mem.duplex == "split":
            _mem.txfreq = mem.offset / 10
        elif mem.duplex == "+":
            _mem.txfreq = (mem.freq + mem.offset) / 10
        elif mem.duplex == "-":
            _mem.txfreq = (mem.freq - mem.offset) / 10
        else:
            _mem.txfreq = _mem.rxfreq

        (txmode, txval, txpol), (rxmode, rxval, rxpol) = \
            chirp_common.split_tone_encode(mem)

        self._encode_tone(_mem.tx_tone, txmode, txval, txpol)
        self._encode_tone(_mem.rx_tone, rxmode, rxval, rxpol)

        _mem.scanadd = mem.skip == "S"
        _mem.narrow = mem.mode == "NFM"

        _mem.ishighpower = mem.power == self.POWER_LEVELS[1]

        # resetting unknowns, this have to be set by hand
        _mem.unknown0 = 3
        _mem.unknown1 = 1

        for setting in mem.extra:
            setting.apply_to_memobj(_mem)

    def get_settings(self):
        _settings = self._memobj.settings
        basic = RadioSettingGroup("basic", "Basic Settings")
        group = RadioSettings(basic)

        # Squelch
        rs = RadioSettingValueInteger(0, 9, _settings.squelch)
        rset = MemSetting("squelch", "Squelch", rs)
        rset.set_doc("Squelch Level: 0, 1, 2, 3, 4, 5, 6, 7, 8, 9")
        basic.append(rset)

        # Time Out Timer
        rs = RadioSettingValueList(TIMEOUTTIMER_LIST,
                                   current_index=_settings.tot)
        rset = MemSetting("tot", "Time-out Timer", rs)
        rset.set_doc("TX Time-out Timer: Off, 30, 60, 90, 120, 150," +
                     " 180 seconds")
        basic.append(rset)

        # Vox Level
        rs = RadioSettingValueList(OFF1TO9_LIST, current_index=_settings.vox)
        rset = MemSetting("vox", "VOX Level", rs)
        rset.set_doc("VOX Level: Off, 1, 2, 3, 4, 5, 6, 7, 8, 9")
        basic.append(rset)

        # Vox Delay
        rs = RadioSettingValueList(VOXD_LIST,
                                   current_index=_settings.voxd)
        rset = MemSetting("voxd", "Vox Delay", rs)
        rset.set_doc("VOX Delay: 0.5, 1.0, 1.5, 2.0, 2.5, 3.0 seconds")
        basic.append(rset)

        # Scan Mode
        rs = RadioSettingValueList(SCANM_LIST, current_index=_settings.scanm)
        rset = MemSetting("scanm", "Scan Mode", rs)
        rset.set_doc("Scan Mode: Carrier, Time")
        basic.append(rset)

        # Voice Annunciation
        rs = RadioSettingValueList(VOICE_LIST,
                                   current_index=_settings.voice)
        rset = MemSetting("voice", "Voice", rs)
        rset.set_doc("Voice Prompts: Off, English")
        basic.append(rset)

        # Side Key 2 (long)
        rs = RadioSettingValueList(SKEY2_LIST,
                                   current_index=_settings.skey2)
        rset = MemSetting("skey2", "Side Key 2", rs)
        rset.set_doc("Side Key 2 (long press): Off, VOX, Power, Scan")
        basic.append(rset)

        # Code Switch
        rs = RadioSettingValueBoolean(_settings.codesw)
        rset = MemSetting("codesw", "Code Switch", rs)
        rset.set_doc("Code Switch: Off, Enabled")
        basic.append(rset)

        # Battery Save
        rs = RadioSettingValueBoolean(_settings.save)
        rset = MemSetting("save", "Battery Save", rs)
        rset.set_doc("Battery Save: Off, Enabled")
        basic.append(rset)

        # Beep Tone
        rs = RadioSettingValueBoolean(_settings.beep)
        rset = MemSetting("beep", "Beep", rs)
        rset.set_doc("Beep Prompt: Off, Enabled")
        basic.append(rset)

        # VOX Switch
        rs = RadioSettingValueBoolean(_settings.voxs)
        rset = MemSetting("voxs", "VOX Switch", rs)
        rset.set_doc("VOX Switch: Off, Enabled")
        basic.append(rset)

        # Roger
        rs = RadioSettingValueBoolean(_settings.roger)
        rset = MemSetting("roger", "Roger", rs)
        rset.set_doc("Roger: Off, Enabled")
        basic.append(rset)

        return group

    def set_settings(self, settings):
        others = settings.apply_to(self._memobj.settings)
        if others:
            LOG.error('Did not apply %s' % others)

    @classmethod
    def match_model(cls, filedata, filename):
        # This radio has always been post-metadata, so never do
        # old-school detection
        return False


@directory.register
@directory.detected_by(h777.RetevisH777)
class H777V4(H777V4BaseRadio):
    """RETEVIS H777V4"""
    VENDOR = "Retevis"
    MODEL = "H777"
    VARIANT = 'V4'
    IDENT = [b'\x00' * 6, b'\xFF' * 6]

    # SKU #: A9294A (sold as FRS radio but supports full band TX/RX)
    # Serial #: 2412R777XXXXXXX

    # SKU #: A9294C (same as SKU #: A9294A (2 pack))
    # Serial #: 2406R777XXXXXXX

    # SKU #: A9294B (sold as PMR radio but supports full band TX/RX)
    # Serial #: 2412R777XXXXXXX

    # SKU #: A9294D (same as SKU #: A9294B (2 pack))
    # Serial #: 24XXR777XXXXXXX



import os
import unittest
from unittest import mock

from chirp.cli import main


class TestCLI(unittest.TestCase):
    def setUp(self):
        self.stdout_lines = []
        self.patches = []
        self.patches.append(mock.patch('sys.exit'))
        self.patches.append(mock.patch.object(main, 'print',
                                              new=self.fake_print))
        for patch in self.patches:
            patch.start()

        # A path to an image file we can pass to the CLI for testing
        self.testfile = os.path.join(
            os.path.abspath(
                os.path.join(
                    os.path.dirname(__file__), '..', 'images',
                    'Icom_IC-2820H.img')))

    def tearDown(self):
        for patch in self.patches:
            patch.stop()

    def fake_print(self, *a):
        for i in a:
            self.stdout_lines.append(str(i))

    @property
    def stdout(self):
        return '\n'.join(self.stdout_lines)

    def test_cli_simple(self):
        # Super simple, just print the first memory and make sure it
        # works
        args = ['--mmap', self.testfile, '--get-mem', 0]
        main.main(args=args)
        self.assertIn('147.56', self.stdout)



# Copyright 2024:
# * Pavel Moravec, OK2MOP <moravecp.cz@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from chirp import chirp_common, directory
from chirp.settings import RadioSettingGroup, RadioSetting, \
    RadioSettingValueBoolean, RadioSettingValueList
from chirp.drivers import baofeng_uv17Pro, baofeng_uv17

LOG = logging.getLogger(__name__)


@directory.register
class RT620(baofeng_uv17.UV17):
    """Radtel RT-620"""
    VENDOR = "Radtel"
    MODEL = "RT-620"
    _radio_memsize = 0x2efff
    _has_workmode_support = True
    _has_gps = True
    _vfoscan = True

    # Top key is referenced in CPS but does not exist in GPS model
    _has_top_key = False
    _has_pilot_tone = True

    STEPS = [2.5, 5.0, 6.25, 10.0, 12.5, 20.0, 25.0, 50.0]

    # Blocks 16-19 - channels, 24-26 channel names, 4 settings,
    #        6 - secondary settings, identity codes,
    #        2 - calibration, 1 - empty radio param
    BLOCK_O_READ = baofeng_uv17.UV17.BLOCK_O_READ + [2]  # back up calibration

    _airband = (108000000, 135999999)
    _vhf_range = (136000000, 299999999)
    _uhf_range = (300000000, 559999999)
    VALID_BANDS = [_airband, _vhf_range, _uhf_range]
    _fingerprint = b"\x06" + b"PROGR6F"
    _magics2 = [(b"\x06", 1),
                (b"\xFF\xFF\xFF\xFF\x0CPROGR6F", 1),
                (b"\02", 8),
                (b"\x06", 1)]
    CHANNELS = 999
    MODES = ["NFM", "FM", "WFM"]
    POWER_LEVELS = [chirp_common.PowerLevel("Low", watts=1.00),
                    chirp_common.PowerLevel("Mid",  watts=5.00),
                    chirp_common.PowerLevel("High",  watts=10.00)]
    LIST_MODE = ["VFO", "Channel Number",
                 "Frequency + Channel Number", "Name + Channel Number"]
    LIST_BEEP = ["Off", "On"]
    LIST_VOX_LEVEL = baofeng_uv17Pro.LIST_VOX_LEVEL
    LIST_ALARMMODE = baofeng_uv17Pro.LIST_ALARMMODE
    LIST_PILOT_TONE = baofeng_uv17Pro.LIST_PILOT_TONE
    LIST_VOX_DELAY_TIME = ["Off"] + ["%s ms" %
                                     x for x in range(500, 2100, 100)]
    LIST_NOAA = ["NOAA OFF", "NOAA Forecast",
                 "NOAA Alarm", "NOAA Both"]  # NOAA here
    LIST_100ms = ["%s ms" % x for x in range(100, 1100, 100)]
    LIST_AB = ["A", "B"]
    LIST_BRIGHTNESS = ["%i" % x for x in range(1, 6)]
    LIST_MENU_QUIT_TIME = ["Off"] + ["%s sec" % x for x in range(1, 31)]
    LIST_KEYS = ["None", "Monitor", "Sweep", "Scan", "Voltage",
                 "Emergency", "Scrambler", "FM Radio", "Compander"]
    MEM_DEFS = """
    struct channel {
      lbcd rxfreq[4];
      lbcd txfreq[4];
      ul16 rxtone;
      ul16 txtone;
      u8 unknown12_7:1,
         bcl:1,
         pttid:2,
         wide:2,
         lowpower:2;
      u8 unknown13_5:3,
         compander:1,
         unknown_13_2:2,
         scrambler:2;
      u8 scode:4,
         unknown14_1:3,
         scan:1;
      u8 step;
    };
    struct channelname {
      char name[11];
    };
    struct settings {
      u8 powerondistype;
      u8 unknown0[15];
      char boottext1[10];
      u8 unknown1[6];
      char boottext2[10];
      u8 unknown2[22];
      u8 tot;
      u8 voxdlytime;
      u8 vox;
      u8 powersave: 4,
         unknown3:1,
         chedit:1,
         voice: 1,
         voicesw: 1;
      u8 backlight;
      u8 beep:1,
         autolock:1,
         unknown_5b5:1,
         noaa:2,
         tail:1,
         dtmfst:2;
      u8 fminterrupt:1,
         dualstandby:1,
         roger:1,
         alarmmode:2,
         alarmtone:1,
         fmenable:1,
         absel:1;
      u8 tailrevert;
      u8 taildelay;
      u8 tone;
      u8 backlightbr;
      u8 squelch;
      u8 chbsquelch;
      u8 chbdistype;
      u8 chadistype;
      u8 menuquittime;
      u8 sk1short;
      u8 sk1long;
      u8 sk2short;
      u8 sk2long;
      u8 topshort;
      u8 toplong;
      u8 scanmode; // Fake Scanmode to satisfy parent method
      u8 unknown8[5];
    };
    struct ani {
      u8 unknown[5];
      u8 code[5];
    };
    struct pttid {
      u8 code[5];
    };
    struct proglocks {
      u8 writelock; //0x00 - off, 0xA5 - on
      u8 readlock;
      char writepass[8];
      char readpass[8];
      u8 empty[13];
      char unknowncode[6]; //Kill code maybe
    };
    struct gps {
      u8 state;
      u8 tzadd12;
    };
    struct dtmfen {
        u8 delay;
        u8 digit;
        u8 interval;
    };
    """
    MEM_LAYOUT = """
    struct {
      ul16 channels;
      ul16 ch_a;
      ul16 ch_b;
    } chsel;
    #seekto 0x0030;
    struct {
      struct channel mem[252];
    } mem1;
    #seek 0x10;
    struct {
      struct channel mem[255];
    } mem2;
    #seek 0x10;
    struct {
      struct channel mem[255];
    } mem3;
    #seek 0x10;
    struct {
      struct channel mem[237];
    } mem4;
    #seekto 0x7000;
    struct settings settings;
    #seekto 0x0010;
    struct {
      struct channel a;
      struct channel b;
    } vfo;
    #seekto 0x4000;
    struct channelname names1[372];
    #seek 0x4;
    struct channelname names2[372];
    #seek 0x4;
    struct channelname names3[255];
    #seekto 0x7100;
    struct gps gps;
    #seekto 0x7069;
    struct proglocks locks;
    #seekto 0x8000;
    struct pttid pttid[15];
    struct ani ani;
    #seekto 0x8061;
    struct dtmfen dtmf;
    """
    MEM_FORMAT = MEM_DEFS + MEM_LAYOUT
    REMOVE_SETTINGS = ['settings.scanmode']

    def remove_extras(self, settings):
        for element in settings:
            if not isinstance(element, RadioSetting):
                settings.remove(element)
                settings.append(self.remove_extra_items(element))
        settings.sort()

    def remove_extra_items(self, group):
        tmp = RadioSettingGroup(group.get_name(), group.get_shortname())
        for element in group:
            name = element.get_name()
            if not isinstance(element, RadioSetting):
                tmp.append(self.remove_extra_items(element))
            elif name not in self.REMOVE_SETTINGS:
                tmp.append(element)
        return tmp

    def get_settings(self):
        top = super().get_settings()
        _mem = self._memobj
        self.remove_extras(top)
        extra = RadioSettingGroup("extra", "Extra Settings")
        self.get_settings_extra(extra, _mem)
        top.append(extra)
        if self._has_gps:
            gps = RadioSettingGroup("gps", "GPS Settings")
            rs = RadioSetting("gps.state", "GPS Enable",
                              RadioSettingValueBoolean(_mem.gps.state))
            gps.append(rs)
            rs = RadioSetting("gps.tzadd12", "GPS Timezone",
                              RadioSettingValueList(
                                  baofeng_uv17Pro.LIST_GPS_TIMEZONE,
                                  current_index=_mem.gps.tzadd12))
            gps.append(rs)
            top.append(gps)
        return top

    def get_settings_extra(self, extra, _mem):
        rs = RadioSetting("settings.absel", "Selected Channel",
                          RadioSettingValueList(
                              self.LIST_AB, current_index=_mem.settings.absel))
        extra.append(rs)
        if _mem.settings.chbsquelch >= len(self.SQUELCH_LIST):
            val = 0x00
        else:
            val = _mem.settings.chbsquelch
        rs = RadioSetting("settings.chbsquelch", "Channel B Squelch",
                          RadioSettingValueList(
                              self.SQUELCH_LIST, current_index=val))
        extra.append(rs)
        # Settings menu 4
        rs = RadioSetting("settings.vox", "Vox Level",
                          RadioSettingValueList(
                              self.LIST_VOX_LEVEL,
                              current_index=_mem.settings.vox))
        extra.append(rs)
        # Settings menu 5
        rs = RadioSetting("settings.voxdlytime", "Vox Delay Time",
                          RadioSettingValueList(
                              self.LIST_VOX_DELAY_TIME,
                              current_index=_mem.settings.voxdlytime))
        extra.append(rs)
        # Settings menu 12
        rs = RadioSetting("settings.backlightbr", "Backlight Brightnes",
                          RadioSettingValueList(
                              self.LIST_BRIGHTNESS,
                              current_index=_mem.settings.backlightbr))
        extra.append(rs)
        rs = RadioSetting("settings.chedit", "Allow Channel Editing",
                          RadioSettingValueBoolean(_mem.settings.chedit))
        extra.append(rs)
        # Settings menu 22
        rs = RadioSetting("settings.tail", "CTCSS Tail Revert",
                          RadioSettingValueBoolean(_mem.settings.tail))
        extra.append(rs)
        rs = RadioSetting("settings.tailrevert", "Repeater Tail Revert Time",
                          RadioSettingValueList(
                              self.LIST_100ms,
                              current_index=_mem.settings.tailrevert))
        extra.append(rs)
        rs = RadioSetting("settings.taildelay", "Repeater Tail Delay Time",
                          RadioSettingValueList(
                              self.LIST_100ms,
                              current_index=_mem.settings.taildelay))
        extra.append(rs)
        rs = RadioSetting("settings.fminterrupt", "FM radio Interruption",
                          RadioSettingValueBoolean(_mem.settings.fminterrupt))
        extra.append(rs)
        # Settings menu 21
        rs = RadioSetting("settings.alarmmode", "Alarm Mode",
                          RadioSettingValueList(
                              self.LIST_ALARMMODE,
                              current_index=_mem.settings.alarmmode))
        extra.append(rs)
        rs = RadioSetting("settings.alarmtone", "Sound Alarm",
                          RadioSettingValueBoolean(_mem.settings.alarmtone))
        extra.append(rs)
        rs = RadioSetting("settings.menuquittime", "Menu Quit Timer",
                          RadioSettingValueList(
                              self.LIST_MENU_QUIT_TIME,
                              current_index=_mem.settings.menuquittime))
        extra.append(rs)
        if self._has_pilot_tone:
            rs = RadioSetting("settings.tone", "Pilot Tone",
                              RadioSettingValueList(
                                  self.LIST_PILOT_TONE,
                                  current_index=_mem.settings.tone))
            extra.append(rs)
        # Settings menu 27
        if _mem.settings.sk1short >= len(self.LIST_KEYS):
            val = 0x00
        else:
            val = _mem.settings.sk1short
        rs = RadioSetting("settings.sk1short", "Side Key 1 Short Press",
                          RadioSettingValueList(
                              self.LIST_KEYS, current_index=val))
        extra.append(rs)
        # Settings menu 28
        if _mem.settings.sk1long >= len(self.LIST_KEYS):
            val = 0x00
        else:
            val = _mem.settings.sk1long
        rs = RadioSetting("settings.sk1long", "Side Key 1 Long Press",
                          RadioSettingValueList(
                              self.LIST_KEYS, current_index=val))
        extra.append(rs)
        # Settings menu 29
        if _mem.settings.sk2short >= len(self.LIST_KEYS):
            val = 0x00
        else:
            val = _mem.settings.sk2short
        rs = RadioSetting("settings.sk2short", "Side Key 2 Short Press",
                          RadioSettingValueList(
                              self.LIST_KEYS, current_index=val))
        extra.append(rs)
        # Settings menu 30
        if _mem.settings.sk2long >= len(self.LIST_KEYS):
            val = 0x00
        else:
            val = _mem.settings.sk2long
        rs = RadioSetting("settings.sk2long", "Side Key 2 Long Press",
                          RadioSettingValueList(
                              self.LIST_KEYS, current_index=val))
        extra.append(rs)
        # Official CPS sets this, but radio does not have top key
        # Probably the non-GPS version or some other clone uses it so
        # I have included it here when/if it will be needed
        if self._has_top_key:
            if _mem.settings.topshort >= len(self.LIST_KEYS):
                val = 0x00
            else:
                val = _mem.settings.topshort
            rs = RadioSetting("settings.topshort", "Top Key Short Press",
                              RadioSettingValueList(
                                  self.LIST_KEYS, current_index=val))
            extra.append(rs)
            # Settings menu 30
            if _mem.settings.toplong >= len(self.LIST_KEYS):
                val = 0x00
            else:
                val = _mem.settings.toplong
            rs = RadioSetting("settings.toplong", "Top Key Long Press",
                              RadioSettingValueList(
                                  self.LIST_KEYS, current_index=val))
            extra.append(rs)
        # NOAA menu
        rs = RadioSetting("settings.noaa", "NOAA Weather mode",
                          RadioSettingValueList(
                              self.LIST_NOAA,
                              current_index=_mem.settings.noaa))
        extra.append(rs)

    def _get_raw_memory(self, number):
        if number == -1:
            return self._memobj.vfo.b
        elif number == -2:
            return self._memobj.vfo.a
        else:
            return super()._get_raw_memory(number)

    def get_memory(self, number):
        mem = super().get_memory(number)
        _mem = self._get_raw_memory(number)
        if hasattr(mem.extra, "keys") and \
                "compander" not in mem.extra.keys():
            rs = RadioSetting("compander", "Compander",
                              RadioSettingValueBoolean(_mem.compander))
            mem.extra.append(rs)
        mem.tuning_step = self.STEPS[_mem.step] or self.STEPS[0]
        if _mem.wide < len(self.MODES):
            mem.mode = self.MODES[_mem.wide]
        return mem

    def set_memory(self, mem):
        _mem = self._get_raw_memory(mem.number)
        super().set_memory(mem)
        _mem.step = self.STEPS.index(mem.tuning_step)
        _mem.wide = self.MODES.index(mem.mode)
        if _mem.rxtone == 0:
            _mem.rxtone = 0xFFFF
        if _mem.txtone == 0:
            _mem.txtone = 0xFFFF

    def get_features(self):
        rf = super().get_features()
        rf.has_tuning_step = True
        rf.has_bank = False
        return rf



from chirp import chirp_common
from tests import base


class TestCaseBanks(base.DriverTest):
    @base.requires_feature('has_bank_names')
    def test_bank_names(self, testname='T'):
        bm = self.radio.get_bank_model()
        banks = bm.get_mappings()

        for bank in banks:
            self.assertIsInstance(bank.get_name(), str,
                                  'Bank model returned non-string name')
            bank.set_name(testname)

        for bank in bm.get_mappings():
            # We allow truncated storage of the name
            self.assertTrue(
                testname.lower().startswith(bank.get_name().lower()),
                'Bank name %r did not stick after set_name(%r)' % (
                    bank.get_name(), testname))
            # Bank names should not contain trailing whitespace
            self.assertEqual(bank.get_name(), bank.get_name().rstrip(),
                             'Bank stored with trailing whitespace')

    @base.requires_feature('has_bank_names')
    def test_bank_names_toolong(self):
        testname = "Not possibly this long"
        self.test_bank_names(testname)

    @base.requires_feature('has_bank_names')
    def test_bank_names_no_trailing_whitespace(self):
        self.test_bank_names('foo  ')

    @base.requires_feature('has_bank')
    def test_bank_store(self):
        loc = self.rf.memory_bounds[0]
        mem = chirp_common.Memory()
        mem.number = loc
        mem.freq = self.rf.valid_bands[0][0] + 100000

        # Make sure the memory is empty and we create it from scratch
        mem.empty = True
        self.radio.set_memory(mem)

        mem.empty = False
        self.radio.set_memory(mem)

        model = self.radio.get_bank_model()

        if isinstance(model, chirp_common.StaticBankModel):
            # Nothing to test with this type
            return

        # If in your bank model every channel has to be tied to a bank, just
        # add a variable named channelAlwaysHasBank to it and make it True
        try:
            channelAlwaysHasBank = model.channelAlwaysHasBank
        except Exception:
            channelAlwaysHasBank = False

        mem_banks = model.get_memory_mappings(mem)
        if channelAlwaysHasBank:
            self.assertNotEqual(0, len(mem_banks),
                                'Freshly-created memory has no banks '
                                'but it should')
        else:
            self.assertEqual(0, len(mem_banks),
                             'Freshly-created memory has banks '
                             'and should not')

        banks = model.get_mappings()

        model.add_memory_to_mapping(mem, banks[0])
        self.assertIn(banks[0], model.get_memory_mappings(mem),
                      'Memory does not claim bank after add')
        self.assertIn(loc, [x.number
                            for x in model.get_mapping_memories(banks[0])],
                      'Bank does not claim memory after add')

        model.remove_memory_from_mapping(mem, banks[0])
        if not channelAlwaysHasBank:
            self.assertNotIn(banks[0], model.get_memory_mappings(mem),
                             'Memory claims bank after remove')
            self.assertNotIn(loc, [x.number
                                   for x in model.get_mapping_memories(
                                       banks[0])],
                             'Bank claims memory after remove')

        if not channelAlwaysHasBank:
            # FIXME: We really need a standard exception for this, because
            # catching Exception here papers over the likely failures from
            # this going unchecked.
            self.assertRaises(Exception,
                              model.remove_memory_from_mapping, mem, banks[0])

    @base.requires_feature('has_bank_index')
    def test_bank_index(self):
        loc = self.rf.memory_bounds[0]
        mem = chirp_common.Memory()
        mem.number = loc
        mem.freq = self.rf.valid_bands[0][0] + 100000

        self.radio.set_memory(mem)

        model = self.radio.get_bank_model()
        banks = model.get_mappings()
        index_bounds = model.get_index_bounds()

        model.add_memory_to_mapping(mem, banks[0])
        for i in range(*index_bounds):
            model.set_memory_index(mem, banks[0], i)
            self.assertEqual(i, model.get_memory_index(mem, banks[0]),
                             'Bank index not persisted')

        suggested_index = model.get_next_mapping_index(banks[0])
        self.assertIn(suggested_index, list(range(*index_bounds)),
                      'Suggested bank index not in valid range')



import logging
import random
import time
from unittest import mock

from chirp import chirp_common
from chirp import errors
from chirp.wxui import serialtrace
from tests import base

LOG = logging.getLogger(__name__)


class SerialException(Exception):
    pass


class SerialNone(serialtrace.SerialTrace):
    def flush(self):
        pass

    def inWaiting(self):
        return len(self.read(256))

    def read(self, size=None):
        return b""

    def write(self, data):
        if not isinstance(data, bytes):
            raise TypeError('Radio wrote non-bytes to serial')

    def setBaudrate(self, rate):
        pass

    def setTimeout(self, timeout):
        pass

    def setParity(self, parity):
        pass

    def __str__(self):
        return self.__class__.__name__.replace("Serial", "")


class SerialError(SerialNone):
    def read(self, size=None):
        raise SerialException("Foo")

    def write(self, data):
        raise SerialException("Bar")


class SerialGarbage(SerialNone):
    def read(self, size=None):
        if size is None:
            size = random.randint(0, 128)
        buf = []
        for i in range(0, size):
            buf.append(i % 256)
        return bytes(buf)


class SerialShortGarbage(SerialNone):
    def read(self, size=None):
        if size is None:
            size = random.randint(0, 128)
        return b'\x01' * (size - 1)


class TestCaseClone(base.DriverTest):
    def setUp(self):
        super().setUp()
        self.live = isinstance(self.radio, chirp_common.LiveRadio)
        self.clone = isinstance(self.radio, chirp_common.CloneModeRadio)

        if not self.clone and not self.live:
            self.skipTest('Does not support clone')

        real_time = time.time

        def fake_time():
            return real_time() * 1000

        self.patches = []
        self.use_patch(mock.patch('time.sleep'))
        self.use_patch(mock.patch('time.time',
                                  side_effect=fake_time))

    def _test_with_serial(self, serial):
        # The base case sets us up with a file, so re-init with our serial.
        # The radio must not read (or fail) with unexpected/error serial
        # behavior on init.
        LOG.info('Initializing radio with fake serial; Radio should not fail')
        orig_mmap = self.parent._mmap

        try:
            cls = self.RADIO_CLASS.detect_from_serial(serial)
            if cls and cls != self.RADIO_CLASS:
                self.fail('Radio detection did not return self')
        except NotImplementedError:
            pass
        except errors.RadioError:
            pass
        except SerialException:
            pass

        self.radio = self.RADIO_CLASS(serial)
        self.radio._mmap = orig_mmap
        self.radio.status_fn = lambda s: True

        msg = ('Clone should have failed and raised an exception '
               'that inherits from RadioError')
        with self.assertRaises(errors.RadioError, msg=msg):
            self.radio.sync_in()

        msg = ('Clone should have failed and raised an exception '
               'that inherits from RadioError')
        with self.assertRaises(errors.RadioError, msg=msg):
            self.radio.sync_out()

    def test_clone_serial_error(self):
        self._test_with_serial(SerialError())

    def test_clone_serial_none(self):
        self._test_with_serial(SerialNone())

    def test_clone_serial_garbage(self):
        self._test_with_serial(SerialGarbage())

    def test_clone_serial_short_garbage(self):
        self._test_with_serial(SerialShortGarbage())
